'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-26 19:42:24
* LastEditors: LiuFeng
* LastEditTime: 2024-05-22 09:42:26
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
from typing import Optional,Union,List
import os
import math
import torch
import numpy as np
from tqdm import tqdm
from ADFWI.model       import AbstractModel
from ADFWI.propagator  import AcousticPropagator,GradProcessor
from ADFWI.survey      import SeismicData
from ADFWI.fwi.misfit  import Misfit,Misfit_NIM
from ADFWI.fwi.regularization import Regularization
from ADFWI.fwi.optimizer import NLCG
from ADFWI.utils       import numpy2tensor
from ADFWI.view        import plot_model

from ADFWI.utils.first_arrivel_picking import apply_mute
from ADFWI.utils.offset_mute import mute_offset
from ADFWI.fwi.multiScaleProcessing import lpass
    

class AcousticFWI(torch.nn.Module):
    """Acoustic Full waveform inversion class
    """
    def __init__(self,
                 propagator:AcousticPropagator,model:AbstractModel,
                 optimizer:torch.optim.Optimizer,scheduler:torch.optim.lr_scheduler,
                 loss_fn:Union[Misfit,torch.autograd.Function],
                 obs_data:SeismicData,
                 gradient_processor: Union[GradProcessor,List[GradProcessor]] = None,
                 regularization_fn:Optional[Regularization]                   = None, 
                 regularization_weights_x:Optional[List[Union[float]]]        = [0,0], # vp/rho in x direction
                 regularization_weights_z:Optional[List[Union[float]]]        = [0,0], # vp/rho in z direction
                 waveform_normalize:Optional[bool]                            = True,
                 waveform_mute_late_window:Optional[float]                    = None,
                 waveform_mute_offset:Optional[float]                         = None,
                 cache_result:Optional[bool]                                  = True,
                 save_fig_epoch:Optional[int]                                 = -1,
                 save_fig_path:Optional[str]                                  = "",
                ):
        """
        Description:
        --------------
        Acoustic Full Waveform Inversion Class
        
        Parameters:
        --------------
        propagator (AcousticPropagator)                                : The propagator used for simulating acoustic wave propagation.
        model (AbstractModel)                                          : The model class representing the velocity or acoustic property structure.
        optimizer (torch.optim.Optimizer)                              : The optimizer used for parameter optimization (e.g., SGD, Adam).
        scheduler (torch.optim.lr_scheduler)                           : The learning rate scheduler for adjusting the learning rate during training.
        loss_fn (Union[Misfit, torch.autograd.Function])               : The loss function or misfit function used to compute the difference between predicted and observed data.
        obs_data (SeismicData)                                         : The observed seismic data for comparison against the model predictions.
        gradient_processor (Union[GradProcessor, List[GradProcessor]]) : The gradient processor or list of processors for handling gradients, applied to different parameters if specified.
        regularization_fn (Optional[Regularization])                   : The regularization function for model parameters (e.g., for smoothing or penalty terms). Default is None.
        regularization_weights_x (Optional[List[Union[float]]])        : Regularization weights for the x direction (e.g., vp/rho regularization). Default is [0, 0].
        regularization_weights_z (Optional[List[Union[float]]])        : Regularization weights for the z direction (e.g., vp/rho regularization). Default is [0, 0].
        waveform_normalize (Optional[bool])                            : Whether to normalize the waveform during inversion. Default is True (waveforms are normalized).
        waveform_mute_late_window:Optional[float]                      : Clipping data after picking the first arrival with the given window size.
        waveform_mute_offset:Optional[float]                           : Clipping data larger than the given offset threshold.
        cache_result (Optional[bool])                                  : Whether to cache intermediate inversion results for later use. Default is True.
        save_fig_epoch (Optional[int])                                 : The interval (in epochs) at which to save the inversion result as a figure. Default is -1 (no figure saved).
        save_fig_path (Optional[str])                                  : The path where to save the inversion result figure. Default is an empty string (no path specified).
        """
        super().__init__()
        self.propagator                 = propagator
        self.model                      = model
        self.optimizer                  = optimizer
        self.scheduler                  = scheduler
        self.loss_fn                    = loss_fn
        self.regularization_fn          = regularization_fn
        self.regularization_weights_x   = regularization_weights_x
        self.regularization_weights_z   = regularization_weights_z
        self.obs_data                   = obs_data
        self.gradient_processor         = gradient_processor
        self.device                     = self.propagator.device
        self.dtype                      = self.propagator.dtype 
        
        # Real-Case settings: for trace missing, partial data missing
        receiver_masks = self.propagator.receiver_masks
        if receiver_masks is None:
            receiver_masks  = np.ones((self.propagator.src_n,self.propagator.rcv_n))
        receiver_masks      = numpy2tensor(receiver_masks)
        self.receiver_masks_2D = receiver_masks # [shot, rcv]
        self.receiver_masks_3D = receiver_masks.unsqueeze(1).expand(-1, self.propagator.nt, -1).to(self.device)  # [shot, time, rcv]
        
        # Real-Case settings: mute late window (by first arrival picking) & mute offset
        self.waveform_normalize         = waveform_normalize
        self.waveform_mute_late_window  = waveform_mute_late_window 
        self.waveform_mute_offset       = waveform_mute_offset
        
        # observed data
        obs_p   = self.obs_data.data["p"]
        obs_p   = numpy2tensor(obs_p,self.dtype).to(self.device)
        if self.propagator.receiver_masks_obs: # mark the observed data need to be masked or not
            obs_p   = obs_p*self.receiver_masks_3D
        self.data_masks = numpy2tensor(self.obs_data.data_masks).to(self.device) if self.obs_data.data_masks is not None else None
        if self.data_masks is not None: # some of the data are unuseful
            obs_p = obs_p*self.data_masks
        self.obs_p = obs_p
        
        # model boundary
        vp_bound =  self.model.get_bound("vp")
        if vp_bound[0] is None and vp_bound[1] is None:
            self.vp_min = self.model.get_model("vp").min() - 500
            self.vp_max = self.model.get_model("vp").max() + 500
        else: 
            self.vp_min = vp_bound[0]
            self.vp_max = vp_bound[1]
            if self.model.water_layer_mask is not None:
                self.vp_min = 1500

        rho_bound =  self.model.get_bound("rho")
        if rho_bound[0] is None and rho_bound[1] is None:
            self.rho_min = self.model.get_model("rho").min() - 500
            self.rho_max = self.model.get_model("rho").max() + 500
        else: 
            self.rho_min = rho_bound[0]
            self.rho_max = rho_bound[1]
            if self.model.water_layer_mask is not None:
                self.rho_min = 1000
        
        # result saving
        self.cache_result   = cache_result
        self.iter_vp, self.iter_rho = [],[]
        self.iter_vp_grad, self.iter_rho_grad = [],[]
        self.iter_loss      = []
        
        # figure saving
        self.save_fig_epoch = save_fig_epoch
        self.save_fig_path  = save_fig_path
    
    def _normalize(self,data):
        mask    = torch.sum(torch.abs(data),axis=1,keepdim=True) == 0
        max_val = torch.max(torch.abs(data),axis=1,keepdim=True).values
        max_val = max_val.masked_fill(mask, 1)
        data = data/max_val
        return data
    
    # misfits calculation
    def calculate_loss(self, synthetic_waveform, observed_waveform, normalization, loss_fn, cutoff_freq=None, propagator_dt=None,shot_index=None):
        """
        Generalized function to calculate misfit loss for a given component.
        Real-Data Processing
            (1) first arrival picking
            (2) mute data by first arrival & giving window
            (3) mute data by offset
            (4) low-pass filter
            (5) data normalize
        """
        # mute data by offset
        if self.waveform_mute_offset is not None:
            receiver_mask_2D = self.receiver_masks_2D[shot_index].cpu() # [shot, rcv]
            src_x            = self.propagator.src_x.cpu()[shot_index]
            rcv_x_list       = self.propagator.rcv_x.cpu()
            rcv_x = torch.zeros(synthetic_waveform.shape[0],synthetic_waveform.shape[-1])
            for i in range(synthetic_waveform.shape[0]):
                rcv_x[i] = rcv_x_list[np.argwhere(receiver_mask_2D[i]).tolist()].squeeze()   
            synthetic_waveform = mute_offset(rcv_x,src_x,self.propagator.dx,synthetic_waveform,self.waveform_mute_offset)
            observed_waveform  = mute_offset(rcv_x,src_x,self.propagator.dx,observed_waveform,self.waveform_mute_offset)
        
        # mute data by first arrival & late window
        if self.waveform_mute_late_window is not None:
            synthetic_waveform_temp = synthetic_waveform.clone()
            observed_waveform_temp  = observed_waveform.clone()
            for i in range(synthetic_waveform.shape[0]):
                synthetic_waveform[i] = apply_mute(self.waveform_mute_late_window, synthetic_waveform_temp[i], self.propagator.dt)
                observed_waveform[i]  = apply_mute(self.waveform_mute_late_window, observed_waveform_temp[i], self.propagator.dt)
        
        # Apply low-pass filter if cutoff frequency is provided
        if cutoff_freq is not None:
            synthetic_waveform, observed_waveform = lpass(synthetic_waveform, observed_waveform, cutoff_freq, int(1 / propagator_dt))
        
        if normalization:
            synthetic_waveform = self._normalize(synthetic_waveform)
            observed_waveform  = self._normalize(observed_waveform)
        
        if isinstance(loss_fn, Misfit):
            return loss_fn.forward(synthetic_waveform, observed_waveform)
        elif isinstance(loss_fn,Misfit_NIM):
            return loss_fn.apply(synthetic_waveform,observed_waveform,loss_fn.p,loss_fn.trans_type,loss_fn.theta)
        else:
            return loss_fn.apply(synthetic_waveform, observed_waveform)
        
    
    # regularization calculation
    def calculate_regularization_loss(self, model_param, weight_x, weight_z, regularization_fn):
        """
        Generalized function to calculate regularization loss for a given parameter.
        """
        regularization_loss = torch.tensor(0.0, device=model_param.device)
        # Check if the parameter requires gradient
        if model_param.requires_grad:
            # Set the regularization weights for x and z directions
            regularization_fn.alphax = weight_x
            regularization_fn.alphaz = weight_z
            # Calculate regularization loss if any weight is greater than zero
            if regularization_fn.alphax > 0 or regularization_fn.alphaz > 0:
                regularization_loss = regularization_fn.forward(model_param)
        return regularization_loss

    # gradient precondition
    def process_gradient(self, parameter,forw,idx=None):
        with torch.no_grad():
            grads = parameter.grad.cpu().detach().numpy()
            vmax = np.max(parameter.cpu().detach().numpy())
            # Apply gradient processor
            if isinstance(self.gradient_processor, GradProcessor):
                grads = self.gradient_processor.forward(nz=self.model.nz, nx=self.model.nx, vmax=vmax, grad=grads, forw=forw)
            else:
                grads = self.gradient_processor[idx].forward(nz=self.model.nz, nx=self.model.nx, vmax=vmax, grad=grads, forw=forw)
            # Convert grads back to tensor and assign
            grads_tensor = numpy2tensor(grads, dtype=self.propagator.dtype).to(self.propagator.device)
            parameter.grad = grads_tensor

    def save_figure(self,i,data,model_type="vp"):
        if self.save_fig_epoch == -1:
            pass
        elif i%self.save_fig_epoch == 0:
            if os.path.exists(self.save_fig_path):
                if model_type == "vp":
                    plot_model(data,title=f"Iteration {i}",
                            dx=self.model.dx,dz=self.model.dz,
                            vmin=self.vp_min,vmax=self.vp_max,
                            save_path=os.path.join(self.save_fig_path,f"{model_type}_{i}.png"),show=False)
                elif model_type == "rho":
                    plot_model(data,title=f"Iteration {i}",
                            dx=self.model.dx,dz=self.model.dz,
                            vmin=self.rho_min,vmax=self.rho_max,
                            save_path=os.path.join(self.save_fig_path,f"{model_type}_{i}.png"),show=False)
                else:
                    plot_model(data,title=f"Iteration {i}",
                            dx=self.model.dx,dz=self.model.dz,
                            save_path=os.path.join(self.save_fig_path,f"{model_type}_{i}.png"),show=False,cmap='coolwarm')
        return
    
    def forward(self,
                iteration:int,
                batch_size:Optional[int]            = None,
                checkpoint_segments:Optional[int]   = 1 ,
                start_iter                          = 0,
                cutoff_freq                         = None,
                ):
        """
        Parameters:
        ------------
        iteration (int)                     : The maximum iteration number in the inversion process.
        batch_size (Optional[int])          : The number of shots (data samples) in each batch. Default is None, meaning use all available shots.
        checkpoint_segments (Optional[int]) : The number of segments into which the time series should be divided for memory efficiency. Default is 1, which means no segmentation.
        start_iter (int)                    : The starting iteration for the optimization process (e.g., for optimizers like Adam/AdamW, and learning rate schedulers like step_lr). Default is 0.
        cutoff_freq (Optional[float])       : The cutoff frequency for low-pass filtering, if specified. Default is None (no filtering applied).
        """
        if isinstance(self.optimizer,torch.optim.LBFGS) or isinstance(self.optimizer,NLCG):
            return self.forward_closure(iteration=iteration,batch_size=batch_size,checkpoint_segments=checkpoint_segments,start_iter=start_iter,cutoff_freq=cutoff_freq)

        n_shots = self.propagator.src_n
        if batch_size is None or batch_size > n_shots:
            batch_size = n_shots

        # epoch
        pbar_epoch = tqdm(range(start_iter,start_iter+iteration),position=0,leave=False,colour='green',ncols=80)
        for i in pbar_epoch:
            # batch
            self.optimizer.zero_grad()
            loss_batch = 0
            pbar_batch = tqdm(range(math.ceil(n_shots/batch_size)),position=1,leave=False,colour='red',ncols=80)
            for batch in pbar_batch:
                # forward simulation
                begin_index = 0  if batch==0 else batch*batch_size
                end_index   = n_shots if batch==math.ceil(n_shots/batch_size)-1 else (batch+1)*batch_size
                shot_index  = np.arange(begin_index,end_index)
                record_waveform = self.propagator.forward(shot_index=shot_index,checkpoint_segments=checkpoint_segments)
                rcv_p,rcv_u,rcv_w = record_waveform["p"],record_waveform["u"],record_waveform["w"]
                forward_wavefield_p,forward_wavefield_u,forward_wavefield_w = record_waveform["forward_wavefield_p"],record_waveform["forward_wavefield_u"],record_waveform["forward_wavefield_w"]
                if batch == 0:
                    forw  = forward_wavefield_p.cpu().detach().numpy()
                else:
                    forw += forward_wavefield_p.cpu().detach().numpy()
                
                # misfit
                if rcv_p.shape == self.obs_p[shot_index].shape: # observed and synthetic data with the same shape (partial-data missing)
                    receiver_mask_3D = self.receiver_masks_3D[shot_index] # [shot, time, rcv]
                    syn_p = rcv_p*receiver_mask_3D 
                else: # observed and synthetic data with the different shape (trace missing)
                    receiver_mask_2D = self.receiver_masks_2D[shot_index] # [shot, rcv]
                    syn_p = torch.zeros_like(self.obs_p[shot_index],device=self.device)
                    for k in range(rcv_p.shape[0]):
                        syn_p[k] = rcv_p[k,...,np.argwhere(receiver_mask_2D[k]).tolist()].squeeze()
                if self.data_masks is not None:
                    data_mask = self.data_masks[shot_index]
                    syn_p = syn_p * data_mask
                data_loss = self.calculate_loss(syn_p, self.obs_p[shot_index], self.waveform_normalize, self.loss_fn, cutoff_freq, self.propagator.dt,shot_index)
                
                # regularization
                if self.regularization_fn is not None:
                    regularization_loss_vp  = self.calculate_regularization_loss(self.model.vp , self.regularization_weights_x[0], self.regularization_weights_z[0], self.regularization_fn)
                    regularization_loss_rho = self.calculate_regularization_loss(self.model.rho, self.regularization_weights_x[1], self.regularization_weights_z[1], self.regularization_fn)
                    regularization_loss = regularization_loss_vp+regularization_loss_rho
                    loss_batch = loss_batch + data_loss.item() + regularization_loss.item()
                    loss = data_loss + regularization_loss
                else:
                    loss_batch = loss_batch + data_loss.item()
                    loss = data_loss
                loss.backward()
                if math.ceil(n_shots/batch_size) == 1:
                    pbar_batch.set_description(f"Shot:{begin_index} to {end_index}")
            
            # gradient process
            if self.model.get_requires_grad("vp"):
                self.process_gradient(self.model.vp, forw=forw, idx=0)
            if self.model.get_requires_grad("rho"):
                self.process_gradient(self.model.rho, forw=forw, idx=1)
        
            self.optimizer.step()
            self.scheduler.step()
            
            # constrain the velocity model
            self.model.forward()
            
            if self.cache_result:
                # model
                temp_vp   = self.model.vp.cpu().detach().numpy()
                temp_rho  = self.model.rho.cpu().detach().numpy()
                self.iter_vp.append(temp_vp)
                self.iter_rho.append(temp_rho)
                self.iter_loss.append(loss_batch)
                self.save_figure(i,temp_vp     , model_type="vp")
                self.save_figure(i,temp_rho    , model_type="rho")
                # gradient
                if self.model.get_requires_grad("vp"):
                    grads_vp   = self.model.vp.grad.cpu().detach().numpy()
                    self.save_figure(i,grads_vp    , model_type="grad_vp")
                    self.iter_vp_grad.append(grads_vp)
                if self.model.get_requires_grad("rho"):
                    grads_rho  = self.model.rho.grad.cpu().detach().numpy()
                    self.save_figure(i,grads_rho   , model_type="grad_rho")
                    self.iter_rho_grad.append(grads_rho)

            self.true_epoch = 0
            pbar_epoch.set_description("Iter:{},Loss:{:.4}".format(i+1,loss_batch))
    
    def forward_closure(self,
                iteration:int,
                batch_size:Optional[int]            = None,
                checkpoint_segments:Optional[int]   = 1 ,
                start_iter                          = 0 ,
                cutoff_freq                         = None,
                ):
        """ inversion using closure version ==> LBFGS,NLCG
        """
        n_shots = self.propagator.src_n
        if batch_size is None or batch_size > n_shots:
            batch_size = n_shots
                
        # epoch
        pbar_epoch = tqdm(range(start_iter,start_iter+iteration),position=0,leave=False,colour='green',ncols=80)
        self.true_epoch = 0
        self.forw = None
        for i in pbar_epoch:
            def closure():
                # batch (for the clouser we hold 1 batch)
                self.optimizer.zero_grad()
                loss_batch = 0
                pbar_batch = tqdm(range(math.ceil(n_shots/batch_size)),position=1,leave=False,colour='red',ncols=80)
                for batch in pbar_batch:
                    # forward simulation
                    begin_index = 0  if batch==0 else batch*batch_size
                    end_index   = n_shots if batch==math.ceil(n_shots/batch_size)-1 else (batch+1)*batch_size
                    shot_index  = np.arange(begin_index,end_index)
                    record_waveform = self.propagator.forward(shot_index=shot_index,checkpoint_segments=checkpoint_segments)
                    rcv_p,rcv_u,rcv_w = record_waveform["p"],record_waveform["u"],record_waveform["w"]
                    forward_wavefield_p,forward_wavefield_u,forward_wavefield_w = record_waveform["forward_wavefield_p"],record_waveform["forward_wavefield_u"],record_waveform["forward_wavefield_w"]
                    if batch == 0:
                        self.forw  = forward_wavefield_p.cpu().detach().numpy()
                    else:
                        self.forw += forward_wavefield_p.cpu().detach().numpy()
                    
                    # misfit
                    if rcv_p.shape == self.obs_p[shot_index].shape: # observed and synthetic data with the same shape (partial-data missing)
                        receiver_mask_3D = self.receiver_masks_3D[shot_index] # [shot, time, rcv]
                        syn_p = rcv_p*receiver_mask_3D 
                    else: # observed and synthetic data with the different shape (trace missing)
                        receiver_mask_2D = self.receiver_masks_2D[shot_index] # [shot, rcv]
                        syn_p = torch.zeros_like(self.obs_p[shot_index],device=self.device)
                        for k in range(rcv_p.shape[0]):
                            syn_p[k] = rcv_p[k,...,np.argwhere(receiver_mask_2D[k]).tolist()].squeeze()
                    if self.data_masks is not None:
                        data_mask = self.data_masks[shot_index]
                        syn_p = syn_p * data_mask
                    data_loss = self.calculate_loss(syn_p, self.obs_p[shot_index], self.waveform_normalize, self.loss_fn, cutoff_freq, self.propagator.dt, shot_index)
                    
                    # regularization
                    if self.regularization_fn is not None:
                        regularization_loss_vp  = self.calculate_regularization_loss(self.model.vp , self.regularization_weights_x[0], self.regularization_weights_z[0], self.regularization_fn)
                        regularization_loss_rho = self.calculate_regularization_loss(self.model.rho, self.regularization_weights_x[1], self.regularization_weights_z[1], self.regularization_fn)
                        regularization_loss = regularization_loss_vp+regularization_loss_rho
                        loss_batch = loss_batch + data_loss.item() + regularization_loss.item()
                        loss = data_loss + regularization_loss
                    else:
                        loss_batch = loss_batch + data_loss.item()
                        loss = data_loss
                    loss.backward()
                    if math.ceil(n_shots/batch_size) == 1:
                        pbar_batch.set_description(f"Shot:{begin_index} to {end_index}")
                self.true_epoch = self.true_epoch + 1
                # gradient process
                if self.model.get_requires_grad("vp"):
                    self.process_gradient(self.model.vp, forw=self.forw, idx=0)
                if self.model.get_requires_grad("rho"):
                    self.process_gradient(self.model.rho, forw=self.forw, idx=1)
                return loss_batch
            
            loss_batch = self.optimizer.step(closure=closure)
            self.scheduler.step()
            
            # constrain the velocity model
            self.model.forward()
            
            # save the result
            if self.cache_result:
                # save the inverted resutls
                temp_vp   = self.model.vp.cpu().detach().numpy()
                temp_rho  = self.model.rho.cpu().detach().numpy()
                self.iter_vp.append(temp_vp)
                self.iter_rho.append(temp_rho)
                self.iter_loss.append(loss_batch)
                
                self.save_figure(i,temp_vp     , model_type="vp")
                self.save_figure(i,temp_rho    , model_type="rho")
                
                # save the inverted gradient
                if self.model.get_requires_grad("vp"):
                    grads_vp   = self.model.vp.grad.cpu().detach().numpy()
                    self.save_figure(i,grads_vp    , model_type="grad_vp")
                    self.iter_vp_grad.append(grads_vp)
                if self.model.get_requires_grad("rho"):
                    grads_rho  = self.model.rho.grad.cpu().detach().numpy()
                    self.save_figure(i,grads_rho   , model_type="grad_rho")
                    self.iter_rho_grad.append(grads_rho)
            pbar_epoch.set_description("Iter:{},Loss:{:.4}".format(i+1,loss_batch))


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-05 19:03:11
* LastEditors: LiuFeng
* LastEditTime: 2024-05-05 19:23:13
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
from .base import Misfit
import torch


class Misfit_waveform_L1(Misfit):
    '''Waveform L1-norm difference misfit (Tarantola, 1984)

    Parameters:
    -----------
        dt (float)      : Time sampling interval.
        obs (Tensor)    : Observed waveform.
        syn (Tensor)    : Synthetic waveform.
    '''
    def __init__(self, dt=1) -> None:
        super().__init__()
        self.dt = dt

    def forward(self, obs, syn):
        '''Compute the L1-norm waveform misfit between observed and synthetic data.

        Args:
            obs (Tensor): Observed waveform.
            syn (Tensor): Synthetic waveform.
        
        Returns:
            Tensor: L1-norm misfit loss.
        '''
        # Calculate residuals by subtracting synthetic from observed data
        rsd = obs - syn

        # Compute the L1-norm loss by summing the absolute value of residuals, weighted by dt
        loss = torch.sum(torch.abs(rsd * self.dt))
        
        return loss


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-14 09:08:12
* LastEditors: LiuFeng
* LastEditTime: 2024-05-26 10:41:48
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

import torch
import torch.nn as nn
import torch.nn.functional as F

def act(act_fun='LeakyReLU'):
    if isinstance(act_fun, str):
        if act_fun == 'LeakyReLU':
            return nn.LeakyReLU(0.1, inplace=True)
        elif act_fun == 'Swish':
            return nn.SiLU()  # PyTorch equivalent of Swish
        elif act_fun == 'ELU':
            return nn.ELU()
        elif act_fun == 'none':
            return nn.Identity()  # Use Identity instead of empty Sequential
        else:
            raise ValueError(f"Unsupported activation: {act_fun}")
    else:
        return act_fun()

def conv(in_f, out_f, kernel_size=3, stride=1, bias=False, pad='same', downsample_mode='stride'):
    downsampler = None
    if stride != 1 and downsample_mode != 'stride':
        if downsample_mode == 'avg':
            downsampler = nn.AvgPool2d(stride, stride)
        elif downsample_mode == 'max':
            downsampler = nn.MaxPool2d(stride, stride)
        else:
            raise ValueError(f"Unsupported downsample mode: {downsample_mode}")
        stride = 1  # Reset stride after pooling

    convolver = nn.Conv2d(in_f, out_f, kernel_size, stride, padding=pad, bias=bias)

    layers = [convolver]
    if downsampler:
        layers.append(downsampler)
    return nn.Sequential(*layers)

def get_block(num_channel, norm_layer, act_fun):
    layers = [
        nn.Conv2d(num_channel, num_channel, 3, 1, 1, bias=False),
        norm_layer(num_channel, affine=True),
        act(act_fun),
        nn.Conv2d(num_channel, num_channel, 3, 1, 1, bias=False),
        norm_layer(num_channel, affine=True),
    ]
    return nn.Sequential(*layers)

class ResidualSequential(nn.Module):
    def __init__(self, *args):
        super(ResidualSequential, self).__init__()
        self.model = nn.Sequential(*args)

    def forward(self, x):
        out = self.model(x)
        # Ensure consistent dimensions or handle mismatch explicitly in ResidualSequential
        if out.size(2) != x.size(2) or out.size(3) != x.size(3):
            x = F.interpolate(x, size=out.size()[2:], mode='bilinear', align_corners=False)
        return out + x

class ResNet(nn.Module):
    def __init__(self, model_shape,
                 random_state_num=100, 
                 vmin=None, vmax=None, 
                 in_channels=1, out_channels=1, 
                 num_blocks=8, num_channel=32, 
                 need_residual=True, 
                 act_fun='LeakyReLU', 
                 norm_layer=nn.InstanceNorm2d, 
                 pad='same',
                 unit=1000,
                 device="cpu"):
        super(ResNet, self).__init__()
        self.vmin   = vmin
        self.vmax   = vmax
        self.unit   = unit
        self.device = device
        
        # neural network blocks
        self.FNN_in = nn.Sequential(
            nn.Linear(in_features=random_state_num, out_features=in_channels*model_shape[0] * model_shape[1], bias=False),
            nn.Unflatten(0,(-1, in_channels, model_shape[0], model_shape[1])),
            nn.LeakyReLU(0.1)
        )

        # residual blocks
        block_class = ResidualSequential if need_residual else nn.Sequential
        layers = [
            conv(in_channels, num_channel, 3, stride=1, bias=False, pad=pad),
            act(act_fun)
        ]
        
        for _ in range(num_blocks):
            layers.append(block_class(*get_block(num_channel, norm_layer, act_fun)))
        
        layers += [
            nn.Conv2d(num_channel, num_channel, 3, 1, 1),
            norm_layer(num_channel, affine=True),
            conv(num_channel, out_channels, 3, 1, bias=False, pad=pad),
            act(act_fun)
        ]
        
        self.res_model = nn.Sequential(*layers)
        
        torch.manual_seed(1234)
        self.random_latent_vector = torch.rand(random_state_num, device=self.device)

    def forward(self):
        # Pass through the main model
        out = self.FNN_in(self.random_latent_vector)
        out = self.res_model(out)
        out = torch.squeeze(out)
        if self.vmin is not None and self.vmax is not None:
            out = ((self.vmax - self.vmin) * torch.tanh(out) + (self.vmax + self.vmin)) / 2
        out = torch.squeeze(out)*self.unit
        return out

    def eval(self):
        self.model.eval()


from .base  import Misfit
from .L1    import Misfit_waveform_L1
from .L2    import Misfit_waveform_L2
from .SmoothL1 import Misfit_waveform_smoothL1
from .Weighted_L1_L2 import Misfit_weighted_L1_and_L2
from .StudentT import Misfit_waveform_studentT

from .Envelope import Misfit_envelope
from .GlobalCorrelation import Misfit_global_correlation
from .Weci import Misfit_weighted_ECI

from .TravelTime import Misfit_traveltime

from .SoftDTW import Misfit_sdtw

from .WDGC import Misfit_weighted_DTW_GC

from .Wasserstein_sinkhorn import Misfit_wasserstein_sinkhorn
# from .Wasserstein_1 import Misfit_Wasserstein1
# from .Wasserstein_1d import Misfit_wasserstein_1d

from .Normalized_Integration_method import Misfit_NIM



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-14 09:08:12
* LastEditors: LiuFeng
* LastEditTime: 2024-05-26 10:41:48
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
import math

class MobileNetV2FWI(nn.Module):
    def __init__(self, model_shape, init_conv_num=4,inverted_residual_blocks=6, vmin=None, vmax=None, random_state_num=100, device="cpu"):
        """
            model_shape (tuple): the shape of velocity model
            inverted_residual_blocks (int): the number of inverted residual blocks, up to 6
            vmin (float): the minimum velocity of output
            vmax (float): the maximum velocity of output
            device (optional): cpu or cuda
        """
        super(MobileNetV2FWI, self).__init__()
        self.vmin = vmin
        self.vmax = vmax
        self.device = device
        self.model_shape = model_shape
        self.init_conv_num = init_conv_num
        # Initial conv layer
        self.init_conv = nn.Sequential(
            nn.Conv2d(init_conv_num, 32, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(32),
            nn.ReLU6(inplace=True)
        )

        # Dynamically create inverted residual blocks
        self.inverted_residual_blocks = nn.ModuleList()
        inp = 32
        channels = [64, 96, 160, 320, 640, 1280]
        for i in range(min(inverted_residual_blocks, 6)):
            outp = channels[i % len(channels)]
            stride = 2 if i % len(channels) == 0 else 1
            self.inverted_residual_blocks.append(InvertedResidual(inp, outp, stride, expand_ratio=6))
            inp = outp

        # Final conv layer
        self.final_conv = nn.Sequential(
            nn.Conv2d(inp, 1280, kernel_size=1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(1280),
            nn.ReLU6(inplace=True)
        )

        # Final layer to map to the desired output channels
        self.output_conv = nn.Conv2d(1280, 1, kernel_size=1, bias=False)
        
        # Resize output to match model_shape
        self.resize = nn.Upsample(size=model_shape, mode='bilinear', align_corners=True)

        # Latent variable
        self.in_features = random_state_num
        torch.manual_seed(1234)
        self.random_latent_vector = torch.rand(self.in_features).to(self.device)

        # Fully connected layer to reshape the latent vector to match MobileNetV2 input size
        self.fc = nn.Linear(self.in_features, self.init_conv_num * (model_shape[0] // 2) * (model_shape[1] // 2))
        
    def forward(self):
        # Generate initial features from the random latent vector
        out = self.fc(self.random_latent_vector)
        out = out.view(1, self.init_conv_num, self.model_shape[0] // 2, self.model_shape[1] // 2)
        
        # Initial conv layer
        out = self.init_conv(out)

        # Pass through inverted residual blocks
        for block in self.inverted_residual_blocks:
            out = block(out)

        # Final conv layer
        out = self.final_conv(out)
        
        # Final convolution to map to single output channel
        out = self.output_conv(out)
        
        # Resize to desired model shape
        out = self.resize(out)
        
        # Post-process the output
        out = torch.squeeze(out)
        if self.vmin is not None and self.vmax is not None:
            out = ((self.vmax - self.vmin) * torch.tanh(out) + (self.vmax + self.vmin)) / 2
        out = out * 1000
        
        return out

class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, expand_ratio):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers = []
        if expand_ratio != 1:
            # pw
            layers.append(nn.Conv2d(inp, hidden_dim, kernel_size=1, stride=1, padding=0, bias=False))
            layers.append(nn.BatchNorm2d(hidden_dim))
            layers.append(nn.ReLU6(inplace=True))
        # dw
        layers.append(nn.Conv2d(hidden_dim, hidden_dim, kernel_size=3, stride=stride, padding=1, groups=hidden_dim, bias=False))
        layers.append(nn.BatchNorm2d(hidden_dim))
        layers.append(nn.ReLU6(inplace=True))
        # pw-linear
        layers.append(nn.Conv2d(hidden_dim, oup, kernel_size=1, stride=1, padding=0, bias=False))
        layers.append(nn.BatchNorm2d(oup))

        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import os
from IPython.display import HTML

def plot_misfit(iter_loss, save_path="", show=False):
    """
    Plot and save the L2-norm misfits over iterations.

    Args:
        iter_loss (list or np.array): The L2-norm misfits for each iteration.
        save_path (str): The path where the plot image will be saved.
        plot (bool): Flag to control whether to display the plot. Default is True.

    Returns:
        None
    """
    # Create a new figure
    plt.figure(figsize=(8, 6))
    
    # Plot the misfit
    plt.plot(iter_loss, c='k')
    plt.xlabel("Iterations", fontsize=12)
    plt.ylabel("Misfits", fontsize=12)
    plt.tick_params(labelsize=12)
    
    # Save the figure
    if not save_path == "":
        plt.savefig(save_path, bbox_inches='tight', dpi=100)
    
    # If plot is True, display the plot
    if show:
        plt.show()
    else:
        # Close the plot
        plt.close()

def plot_initial_and_inverted(vp_init, iter_vp, save_path="", show=False,cmap='jet_r'):
    """
    Plot and save the initial model and inverted results.

    Args:
        vp_init (np.array): The initial velocity model.
        iter_vp (list or np.array): List of inverted results, with the last element being the final model.
        save_path (str): The path where the plot image will be saved.
        plot (bool): Flag to control whether to display the plot. Default is False.

    Returns:
        None
    """
    # Create a new figure
    plt.figure(figsize=(12, 8))
    
    # Plot initial model
    plt.subplot(121)
    plt.imshow(vp_init, cmap=cmap)
    plt.title("Initial Model")
    
    # Plot inverted results
    plt.subplot(122)
    plt.imshow(iter_vp[-1], cmap=cmap)
    plt.title("Inverted Model")

    # Save the figure
    if save_path:
        plt.savefig(save_path, bbox_inches='tight', dpi=100)
    
    # If plot is True, display the plot
    if show:
        plt.show()
    else:
        # Close the plot
        plt.close()

def animate_inversion_process(iter_vp, vmin=None,vmax=None, save_path="", fps=10, interval=150):
    """
    Create an animation of the inversion process.

    Args:
        iter_vp (list or np.array): List of inverted results over iterations.
        vp_true (np.array): The true velocity model for color scaling.
        save_path (str): The path where the animation will be saved.
        fps (int): Frames per second for the animation. Default is 10.

    Returns:
        None
    """
    # Set up the figure for plotting
    fig, ax = plt.subplots(figsize=(8, 6))
    vmin = vmin if vmin is not None else iter_vp.min()
    vmax = vmax if vmax is not None else iter_vp.max()
    cax = ax.imshow(iter_vp[0], aspect='equal', cmap='jet_r', vmin=vmin, vmax=vmax)
    ax.set_title('Inversion Process Visualization')
    ax.set_xlabel('X Coordinate')
    ax.set_ylabel('Z Coordinate')

    # Create a horizontal colorbar
    cbar = fig.colorbar(cax, ax=ax, orientation='horizontal', fraction=0.046, pad=0.2)
    cbar.set_label('Velocity (m/s)')

    # Adjust the layout to minimize white space
    plt.subplots_adjust(top=0.85, bottom=0.2, left=0.1, right=0.9)

    # Initialization function
    def init():
        cax.set_array(iter_vp[0])  # Use the 2D array directly
        return cax,

    # Animation function
    def animate(i):
        cax.set_array(iter_vp[i])  # Update with the i-th iteration directly
        return cax,

    # Create the animation
    ani = animation.FuncAnimation(fig, animate, init_func=init, frames=len(iter_vp), interval=interval, blit=True)

    # Save the animation as a video file (e.g., GIF format)
    if save_path:
        ani.save(save_path, writer='pillow', fps=fps)

    plt.close(fig)  # Prevents static display of the last frame



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-26 19:42:24
* LastEditors: LiuFeng
* LastEditTime: 2024-05-13 23:16:49
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from ADFWI.model import AbstractModel
from ADFWI.utils import numpy2tensor
from ADFWI.view import plot_vp_rho,plot_model
from typing import Optional,Tuple,Union
import torch
from torch import Tensor
import numpy as np
from torchinfo import summary

class DIP_AcousticModel(AbstractModel):
    """Acoustic Velocity model with deep parameterization of vp or rho
    """
    def __init__(self,
                ox:float,oz:float,
                nx:int  ,nz:int,
                dx:float,dz:float,
                DIP_model_vp                                     = None,     # deep image prior models
                DIP_model_rho                                    = None,
                reparameterization_strategy                      = "vel",    # vel/vel_diff
                reparameterization_inputs                        = "random", # random/vel/update_vel
                vp_init:Optional[Union[np.array,Tensor]]         = None,     # initial model parameter
                rho_init:Optional[Union[np.array,Tensor]]        = None,
                vp_bound    : Optional[Tuple[float, float]]      = None,     # model parameter's boundary
                rho_bound   : Optional[Tuple[float, float]]      = None,
                water_layer_mask:Optional[Union[np.array,Tensor]]= None,
                auto_update_rho:Optional[bool]                   = True,
                auto_update_vp :Optional[bool]                   = False,
                free_surface: Optional[bool]                     = False,
                abc_type    : Optional[str]                      = 'PML',
                abc_jerjan_alpha:Optional[float]                 = 0.0053,
                nabc:Optional[int]                               = 20,
                device                                           = 'cpu',
                dtype                                            = torch.float32
                )->None:
        """
        Parameters:
        --------------
        ox (float), oz (float)                               : Non use, the origin coordinates of the model in the x- and z- directions (in meters).
        nx (int), nz (int)                                   : The number of grid points in the x- and z- directions.
        dx (float), dz (float)                               : The grid spacing in the x- and z- directions (in meters).
        DIP_model_vp                                         : reparameterized vp using a deep neural network, by default None
        DIP_model_rho                                        : reparameterized rho using a deep neural network, by default None
        reparameterization_strategy (str)                    : reparameterization strategy (vel for generating velocity model and vel_diff for generating velocity variation)
        vp_init                                              : the initial vp model (must when use the vel_diff strategy)
        rho_init                                             : the initial rho model (must when use the vel_diff strategy)
        vp_bound (Optional[Tuple[float, float]])             : The lower and upper bounds for the P-wave velocity model. Default is None.
        rho_bound (Optional[Tuple[float, float]])            : The lower and upper bounds for the density model. Default is None.
        auto_update_rho (Optional[bool])                     : Whether to automatically update the density model during inversion. Default is True.
        auto_update_vp (Optional[bool])                      : Whether to automatically update the P-wave velocity model during inversion. Default is False.
        water_layer_mask (Optional[Union[np.array, Tensor]]) : A mask for the water layer (not update), if applicable. Default is None.
        free_surface (Optional[bool])                        : A flag to indicate the presence of a free surface in the model. Default is False.
        abc_type (Optional[str])                             : The type of absorbing boundary condition used in the model. Options include 'PML' and 'Jerjan'. Default is 'PML'.
        abc_jerjan_alpha (Optional[float])                   : The attenuation factor for the Jerjan boundary condition. Default is 0.0053.
        nabc (Optional[int])                                 : The number of grid cells dedicated to the absorbing boundary, default is 20.
        device (str)                                         : The device on which to run the model. Options are 'cpu' or 'cuda'. Default is 'cpu'.
        dtype (torch.dtype)                                  : The data type for PyTorch tensors. Default is torch.float32.
        """
        # initialize the common model parameters
        super().__init__(ox,oz,nx,nz,dx,dz,free_surface,abc_type,abc_jerjan_alpha,nabc,device,dtype)
        self.reparameterization_strategy = reparameterization_strategy
        self.reparameterization_inputs = reparameterization_inputs
        
        # update rho/vp using the empirical function
        self.auto_update_rho = auto_update_rho
        self.auto_update_vp  = auto_update_vp
        
        # gradient mask
        if water_layer_mask is not None:
            self.water_layer_mask = numpy2tensor(water_layer_mask,dtype=torch.bool).to(device)
        else:
            self.water_layer_mask = None
        
        # Neural networks
        self.DIP_model_vp   = DIP_model_vp
        self.DIP_model_rho  = DIP_model_rho
        
        # initialize the model parameters
        self.pars       = ["vp","rho"]
        self.vp_init    = torch.zeros((nz,nx),dtype=dtype).to(device) if  vp_init is None else numpy2tensor(vp_init,dtype=dtype).to(device)
        self.rho_init   = torch.zeros((nz,nx),dtype=dtype).to(device) if rho_init is None else numpy2tensor(rho_init,dtype=dtype).to(device)
        self.vp         = self.vp_init.clone()
        self.rho        = self.rho_init.clone()
        
        if self.reparameterization_inputs == "vel":
            self._parameterization(self.vp_init.unsqueeze(0).unsqueeze(0))
        elif self.reparameterization_inputs == "update_vel":
            self._parameterization(self.vp_init.unsqueeze(0).unsqueeze(0))
        else:
            self._parameterization()
        
        # set model bounds
        self.lower_bound["vp"]  =  vp_bound[0]  if vp_bound  is not None else None
        self.lower_bound["rho"] = rho_bound[0]  if rho_bound is not None else None
        self.upper_bound["vp"]  =  vp_bound[1]  if vp_bound  is not None else None
        self.upper_bound["rho"] = rho_bound[1]  if rho_bound is not None else None
        
        # check the input model
        self._check_bounds()
        self.check_dims()
        
    def get_requires_grad(self, par: str) -> bool:
        """Return the gradient of the model

        Parameters
        ----------
        par (str) : Model parameter name

        Returns
        -------
        grad (bool) : Flag for gradient of the model
        """

        if par not in self.pars:
            raise ValueError("Parameter {} not in model".format(par))
        if par == "vp":
            return self.DIP_model_vp is not None
        if par == "rho":
            return self.DIP_model_rho is not None

    def get_model(self, par: str):
        if par not in ["vp","rho"]:
            raise "Error input parametrs"
        elif par == "vp":
            vp  = self.vp.cpu().detach().numpy()
            return vp
        elif par == "rho":
            rho = self.rho.cpu().detach().numpy()
            return rho

    def get_bound(self, par: str) -> Tuple[float, float]:
        if par not in ["vp","rho"]:
            raise "Error input parameters"
        else:
            m_min = self.lower_bound[par]
            m_max = self.upper_bound[par]        
        return [m_min,m_max]
    
    def __repr__(self) -> str:
        """Representation of the model object

        Returns
        -------
        repr (str) : Representation of the model object
        """
        info = f"   Model with parameters {self.pars}:\n"
        info += f"  Model orig: ox = {self.ox:6.2f}, oz = {self.oz:6.2f} m\n"
        info += f"  Model grid: dx = {self.dx:6.2f}, dz = {self.dz:6.2f} m\n"
        info += f"  Model dims: nx = {self.nx:6d}, nz = {self.nz:6d}\n"
        info += f"  Model size: {self.nx * self.nz * len(self.pars)}\n"
        info += f"  Free surface: {self.free_surface}\n"
        info += f"  Absorbing layers: {self.nabc}\n"
        info += f"  NN structure\n"
        if self.DIP_model_vp is not None:
            info += str(summary(self.DIP_model_vp,device=self.device))
        if self.DIP_model_rho is not None:
            info += str(summary(self.DIP_model_rho,device=self.device))
        return info
    
    def set_rho_using_empirical_function(self):
        """approximate rho via empirical relations with vp
        """
        vp          = self.vp.cpu().detach().numpy()
        rho         = self.rho.cpu().detach().numpy()
        rho_emprical= np.power(vp, 0.25) * 310
        if self.water_layer_mask is not None:
            grad_mask = self.water_layer_mask.cpu().detach().numpy()
            rho_emprical[grad_mask] = rho[grad_mask]
        self.rho    = numpy2tensor(rho_emprical,self.dtype).to(self.device)
        return
    
    def set_vp_using_empirical_function(self):
        """approximate vp via empirical relations with rho
        """
        rho         = self.rho.cpu().detach().numpy()
        vp          = self.vp.cpu().detach().numpy()
        vp_empirical= np.power(rho / 310, 4)
        if self.water_layer_mask is not None:
            grad_mask = self.water_layer_mask.cpu().detach().numpy()
            vp_empirical[grad_mask] = vp[grad_mask]
        self.vp     = numpy2tensor(vp_empirical,self.dtype).to(self.device)
        return   
    
    def _parameterization(self,*args,**kw_args):
        """setting variable and gradients
        """
        if self.DIP_model_vp is not None:
            if self.reparameterization_strategy == "vel":
                self.vp     = self.DIP_model_vp(*args,**kw_args)
            elif self.reparameterization_strategy == "vel_diff":
                self.vp     = self.vp_init + self.DIP_model_vp(*args,**kw_args)
        elif self.auto_update_vp:
            self.set_vp_using_empirical_function()
            
        if self.DIP_model_rho is not None:
            if self.reparameterization_strategy == "vel":
                self.rho    = self.DIP_model_rho(*args,**kw_args)
            elif self.reparameterization_strategy == "vel_diff":
                self.rho    = self.rho_init + self.DIP_model_rho(*args,**kw_args)
        elif self.auto_update_rho:
            self.set_rho_using_empirical_function()
        return
    
    def _plot_vp_rho(self,**kwargs):
        """plot velocity model
        """
        plot_vp_rho(self.vp,self.rho, dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot(self,var,**kwargs):
        """plot single velocity model
        """
        model_data = self.get_model(var)
        plot_model(model_data,title=var,**kwargs)
        return
    
    def clip_params(self,par)->None:
        """Clip the model parameters to the given bounds
        """
        if self.get_requires_grad(par):
            if self.lower_bound[par] is not None and self.upper_bound[par] is not None:
                # Retrieve the model parameter
                m = getattr(self, par)
                min_value = self.lower_bound[par]
                max_value = self.upper_bound[par]
                # Create a temporary copy for masking purposes
                m_temp = m.clone()  # Use .clone() instead of .copy() to avoid issues with gradients

                # Clip the values of the parameter using in-place modification with .data
                m.data.clamp_(min_value, max_value)
                
                # Apply the water layer mask if it is not None, using in-place modification
                if self.water_layer_mask is not None:
                    m.data = torch.where(self.water_layer_mask.contiguous(), m_temp.data, m.data)
        return

    def forward(self,*args,**kwargs) -> Tuple:
        """Forward method of the elastic model class
        """
        if self.reparameterization_inputs == "vel":
            self._parameterization(self.vp_init.unsqueeze(0).unsqueeze(0))
        elif self.reparameterization_inputs == "update_vel":
            self._parameterization(numpy2tensor(self.get_model("vp")).to(self.device).unsqueeze(0).unsqueeze(0))
        else:
            self._parameterization()
        
        self.clip_params("vp")
        self.clip_params("rho")
        return


"""
* Modified from: https://github.com/seisfwi/SWIT
* Original Author: Haipeng Li
* Original Author Email: haipeng@stanford.edu
=========================================================
* Author: Liu Feng (SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: Liu Feng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by Liu Feng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
"""

import numpy as np 
import matplotlib.pyplot as plt

from typing import Optional,List,Union
from ADFWI.survey import Survey
from ADFWI.utils import gpu2cpu,tensor2numpy
from ADFWI.view import plot_waveform2D,plot_waveform_wiggle,plot_waveform_trace

class SeismicData():
    def __init__(self,survey:Survey):
        self.survey     = survey
        # get the survey information
        self.src_num    = survey.source.num
        self.rcv_num    = survey.receiver.num
        self.src_loc    = survey.source.get_loc()
        self.rcv_loc    = survey.receiver.get_loc()
        self.src_type   = survey.source.get_type()
        self.rcv_type   = survey.receiver.get_type()
        self.nt         = survey.receiver.nt
        self.dt         = survey.receiver.dt
        self.t          = np.arange(self.nt)*self.dt
        
        # data
        self.data = None
        self.data_masks = None
        
    def __repr__(self):
        """ Print the survey information
        """

        info = f"Seismic Data:\n"
        info += f"  Source number : {self.src_num}\n"
        info += f"  Receiver number : {self.rcv_num}\n"
        info += f"  Time samples : {self.nt} samples at {self.dt * 1000:.2f} ms\n"

        return info

    def record_data(self, data: dict):
        """ Add the shot gather data to the class

        Parameters:
        ----------
        data: dict
            shot gather data in dictionary format
        """
        for key,value in data.items():
            value      = tensor2numpy(gpu2cpu(value)).copy()
            data[key]  = value
        self.data = data
    
    def save(self,path:str):
        data_save = {   'data'      : self.data,
                        'src_loc'   : self.src_loc,
                        'rcv_loc'   : self.rcv_loc,
                        'src_num'   : self.src_num,
                        'rcv_num'   : self.rcv_num,
                        'rcv_type'  : self.rcv_type,
                        'src_type'  : self.src_type,
                        't'         : self.t,
                        'nt'        : self.nt,
                        'dt'        : self.dt
                    }
        np.savez(path, **data_save) 
    
    def load(self, path: str):
        """ Load the shot gather data

        Parameters:
        ----------
        path: str
            load path
        """

        data = np.load(path, allow_pickle=True)

        # load the data
        self.data       = data['data'].item()
        self.src_loc    = data['src_loc']
        self.rcv_loc    = data['rcv_loc']
        self.src_num    = data['src_num']
        self.rcv_num    = data['rcv_num']
        self.rcv_type   = data['rcv_type']
        self.src_type   = data['src_type']
        self.t          = data['t']
        self.nt         = data['nt']
        self.dt         = data['dt']
        return
    
    def normalize_and_mask(self,array):
        time_sum = np.sum(np.abs(array), axis=1, keepdims=True)
        mask = time_sum == 0
        max_val  = np.max(np.abs(array), axis=1, keepdims=True)
        max_val = np.where(mask, 1, max_val)
        array = array / max_val
        return array
    
    def parse_elastic_data(self,normalize=False):
        txx = self.data["txx"]
        tzz = self.data["tzz"]
        txz = self.data["txz"]
        vx  = self.data["vx"]
        vz  = self.data["vz"]
        pressure = -(txx + tzz)
        if normalize:
            pressure = self.normalize_and_mask(pressure)
            txz      = self.normalize_and_mask(txz)
            vx       = self.normalize_and_mask(vx)
            vz       = self.normalize_and_mask(vz)
        return pressure,txz,vx,vz

    def parse_acoustic_data(self,normalize=False):
        pressure = self.data["p"]
        u = self.data["u"]
        w = self.data["w"]
        if normalize:
            pressure = self.normalize_and_mask(pressure)
            u = self.normalize_and_mask(u)
            w = self.normalize_and_mask(w)

        return pressure,u,w    
    
    def plot_waveform2D(self,i_shot,rcv_type="pressure",acoustic_or_elastic="acoustic",normalize=True,**kwargs):
        if acoustic_or_elastic == "acoustic":
            pressure,vx,vz = self.parse_acoustic_data(normalize=normalize)
        elif acoustic_or_elastic == "elastic":
            pressure,txz,vx,vz = self.parse_elastic_data(normalize=normalize)
        
        if rcv_type     == "pressure":
            plot_waveform2D(pressure[i_shot].T,**kwargs)
        elif rcv_type   == "vx":
            plot_waveform2D(vx[i_shot].T,**kwargs)
        elif rcv_type   == "vz":
            plot_waveform2D(vz[i_shot].T,**kwargs)
        elif rcv_type   == "txz":
            plot_waveform2D(txz[i_shot].T,**kwargs)
        return
    
    def plot_waveform_wiggle(self,i_shot,rcv_type="pressure",acoustic_or_elastic="acoustic",normalize=True,**kwargs):
        if acoustic_or_elastic == "acoustic":
            pressure,vx,vz = self.parse_acoustic_data(normalize=normalize)
        elif acoustic_or_elastic == "elastic":
            pressure,txz,vx,vz = self.parse_elastic_data(normalize=normalize)
        
        if rcv_type == "pressure":
            plot_waveform_wiggle(pressure[i_shot],self.survey.source.t,**kwargs)
        elif rcv_type == "vx":
            plot_waveform_wiggle(vx[i_shot],self.survey.source.t,**kwargs)
        elif rcv_type == "vz":
            plot_waveform_wiggle(vz[i_shot],self.survey.source.t,**kwargs)
        elif rcv_type == "txz":
            plot_waveform_wiggle(txz[i_shot],self.survey.source.t,**kwargs)
        return
    
    def plot_waveform_trace(self,i_shot,i_trace,rcv_type="pressure",acoustic_or_elastic="acoustic",normalize=True,**kwargs):
        if acoustic_or_elastic == "acoustic":
            pressure,vx,vz = self.parse_acoustic_data(normalize=normalize)
        elif acoustic_or_elastic == "elastic":
            pressure,txz,vx,vz = self.parse_elastic_data(normalize=normalize)
        
        if rcv_type == "pressure":
            plot_waveform_trace(pressure,i_shot,i_trace,self.dt,**kwargs)
        elif rcv_type == "vx":
            plot_waveform_trace(vx,i_shot,i_trace,self.dt,**kwargs)
        elif rcv_type == "vz":
            plot_waveform_trace(vz,i_shot,i_trace,self.dt,**kwargs)
        elif rcv_type == "txz":
            plot_waveform_trace(txz,i_shot,i_trace,self.dt,**kwargs)
        return


import numpy as np
from skimage.metrics import structural_similarity as ssim

def MSE(true_v,inv_v):
    nz,nx = true_v.shape
    return np.sum((true_v-inv_v)**2)

def RMSE(true_v,inv_v):
    if len(true_v.shape) != len(inv_v.shape):
        true_v = true_v[np.newaxis,:,:]
        res = np.sqrt(np.sum((true_v-inv_v)**2,axis=(1,2)))
        return res
    else:
        return np.sqrt(np.sum((true_v-inv_v)**2))

def MAPE(true_v,inv_v):
    if len(true_v.shape) != len(inv_v.shape):
        nz,nx = true_v.shape
        true_v = true_v[np.newaxis,:,:]
        res = 100/(nx*nz) * np.sum(np.abs(inv_v-true_v)/true_v,axis=(1,2))
    else:
        nz,nx = true_v.shape[-2:]
        res = 100/(nx*nz) * np.sum(np.abs(inv_v-true_v)/true_v)
    return res

def SSIM(true_v,inv_v,win_size=11):
    if len(true_v.shape) != len(inv_v.shape):
        ssim_res = []
        for i in range(inv_v.shape[0]):
            vmax = np.max([true_v.max(),inv_v[i].max()])
            vmin = np.min([true_v.min(),inv_v[i].min()])
            temp_ssim =  ssim(true_v,inv_v[i],data_range=vmax-vmin,win_size=win_size)
            ssim_res.append(temp_ssim)
        return ssim_res
    else:
        vmax = np.max([true_v.max(),inv_v.max()])
        vmin = np.min([true_v.min(),inv_v.min()])
        return ssim(true_v,inv_v,data_range=vmax-vmin,win_size=win_size)
    
def SNR(true_v,inv_v):
    if len(true_v.shape) != len(inv_v.shape):
        snr_res = []
        for i in range(inv_v.shape[0]):
            true_norm  = np.sum(true_v**2)
            diff_norm = np.sum((true_v-inv_v[i])**2)
            temp_snr = 10*np.log10(true_norm/diff_norm)
            snr_res.append(temp_snr)
        return snr_res
    else:
        true_norm  = np.sum(true_v**2)
        diff_norm = np.sum((true_v-inv_v)**2)
        temp_snr = 10*np.log10(true_norm/diff_norm)
        return temp_snr


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
from .L1 import Misfit_waveform_L1
from .L2 import Misfit_waveform_L2
import math
import torch


class Misfit_weighted_L1_and_L2(Misfit):
    """ Some detail can be find in:
        http://www.sce.carleton.ca/faculty/adler/talks/2013/rahmati-CMBES2013-weighted-L1-L2-pres.pdf

    Paraemters:
    -----------
        dt (float)      : time sampling interval
        max_iter (int)  : the maximum iteration
        obs (Tensors)   : the observed waveform 
        syn (Tensors)   : the synthetic waveform 
    """
    def __init__(self,dt=1,max_iter=1000) -> None:
        super().__init__()
        self.dt     = dt
        self.iter   = 0
        self.max_iter = max_iter
        self.L1_fn = Misfit_waveform_L1(dt=self.dt)
        self.L2_fn = Misfit_waveform_L2(dt=self.dt)
        
    def forward(self,obs,syn):
        N = self.max_iter
        w_i = 1/(1+math.exp(-(self.iter - N/2)))
        L1_loss = self.L1_fn.forward(obs=obs,syn=syn)
        L2_loss = self.L2_fn.forward(obs=obs,syn=syn)
        loss = w_i*L2_loss + (1-w_i)*L1_loss
        self.iter += 1
        return loss


import numpy as np
from ADFWI.utils.utils import tensor2numpy
import torch

def mute_offset(rcv_x,src_x,dx,waveform,distance_threshold = 400):
    """ mute from the shot to each side
        unit = m
    """
    rcv_x = tensor2numpy(rcv_x)
    src_x = tensor2numpy(src_x)
    for ishot in range(waveform.shape[0]):
        distance_mask = np.abs(rcv_x[ishot] - src_x[ishot]) < distance_threshold/dx
        distance_mask = ~distance_mask # [rcv]
        distance_mask_temp = torch.zeros_like(waveform[ishot]).to(waveform.device)
        distance_mask_temp[:,distance_mask.tolist()] = 1
        waveform[ishot] = waveform[ishot]*distance_mask_temp
    return waveform
        
    


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-12 09:09:07
* LastEditors: LiuFeng
* LastEditTime: 2024-05-12 09:09:14
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
from .base import Misfit
import torch
import math
from typing import Optional
from .GlobalCorrelation import Misfit_global_correlation
from .SoftDTW import Misfit_sdtw

class Misfit_weighted_DTW_GC(Misfit):
    """Weighted softdtw and global correlation misfit (Song Chao et al., 2023, IEEE TGRS)
    
    Paraemters:
    -----------
        dt (float)      : time sampling interval
        gamma (float)           : Regularization parameter. It should be strictly positive. Lower is less smoothed (closer to true DTW).
        sparse_sampling (int)   : down-sampling the signal for accelerating inversion
        obs (Tensors)           : the observed waveform [shot,time sampling,receiver]
        syn (Tensors)           : the synthetic waveform [shot,time sampling,receiver]

    """
    def __init__(self,max_iter=1000,gamma:Optional[float]=1,sparse_sampling:Optional[int]=1,dt:Optional[float]=1) -> None:
        super().__init__()
        self.max_iter = max_iter
        self.iter = 0
        self.dt = dt
        self.GC_fn       = Misfit_global_correlation(dt=self.dt)
        self.Softdtw_fn  = Misfit_sdtw(gamma=gamma,sparse_sampling=sparse_sampling,dt=dt)
        
    def forward(self,obs,syn):
        N = self.max_iter
        w_i = 1/(1+math.exp(-(self.iter - N/2)))
        GCN_loss = self.GC_fn.forward(obs=obs,syn=syn)
        DTW_loss = self.Softdtw_fn.forward(obs=obs,syn=syn)
        loss = w_i*GCN_loss + (1-w_i)*DTW_loss
        self.iter += 1
        return loss


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-14 09:08:12
* LastEditors: LiuFeng
* LastEditTime: 2024-05-26 10:41:48
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

import torch
import torch.nn as nn
import torch.nn.functional as F
import math

############################################################################
#                   CNN model implement by Weiqiang Zhu
# Zhu, W., Xu, K., Darve, E. & Beroza, G.C., 2021. 
#   A general approach to seismic inversion with automatic differentiation. 
#   Computers & Geosciences, 151, 104751. 
#   doi:10.1016/j.cageo.2021.104751
# Zhu, W., Xu, K., Darve, E., Biondi, B. & Beroza, G.C., 2022. 
#   Integrating deep neural networks with full-waveform inversion: Reparameterization, regularization, and uncertainty quantification. 
#   GEOPHYSICS, 87, R93–R109. 
#   doi:10.1190/geo2020-0933.1  
############################################################################
# model
class CNN(torch.nn.Module):
    def __init__(self,model_shape,
                 in_channels  = [8,32,16],
                 vmin=None,vmax=None,
                 dropout_prob=0,
                 random_state_num = 100,
                 unit = 1000,
                 device="cpu"):
        """
            model_shape (tuple) : the shape of velocity model
            in_channels (list)  : the input and output channels for each CNN block
            vmin (float)        : the minimum velocity of output
            vmax (float)        : the maximum velocity of output
            unit                : the unit of the model parameters
            dropout_prob (float): probability for dropout
            random_state_num    : the input parameters of network
            device (optional)   : cpu or cuda
        """
        super(CNN,self).__init__()
        self.vmin = vmin
        self.vmax = vmax
        self.device = device
        self.unit = unit
        
        # model setting
        self.layer_num  = layer_num = len(in_channels)-1
        h_in            = math.ceil(model_shape[0]/(2**layer_num))
        w_in            = math.ceil(model_shape[1]/(2**layer_num))
        self.h_v0       = model_shape[0]
        self.w_v0       = model_shape[1]
        
        # neural network blocks
        self.in_features = random_state_num
        
        self.FNN_in = nn.Sequential(
            nn.Linear(in_features=self.in_features,out_features=h_in*w_in*in_channels[0],bias=False),
            nn.Unflatten(0,(-1,in_channels[0],h_in,w_in)),
            nn.LeakyReLU(0.1)
        )
        
        self.CNN_Blocks = nn.ModuleList()
        for i in range(layer_num):
            self.CNN_Blocks.append(
                nn.Sequential(
                    nn.UpsamplingBilinear2d(scale_factor=(2,2)),
                    nn.Conv2d(in_channels = in_channels[i],out_channels=in_channels[i+1],kernel_size=4,stride=1,padding="same",bias=False),
                    nn.LeakyReLU(0.1),
                    nn.Dropout(p=dropout_prob)  # add dropout layer
                )
            )
        
        self.CNN_out = nn.Sequential(
            nn.Conv2d(in_channels = in_channels[-1],out_channels=1,kernel_size=4,stride=1,padding="same",bias=False)
        )
        
        # latent variable
        torch.manual_seed(1234)
        self.random_latent_vector = torch.rand(self.in_features).to(self.device)
    
    def forward(self):
        # neural network generation
        out = self.FNN_in(self.random_latent_vector)
        for i in range(self.layer_num):
            out = self.CNN_Blocks[i](out)
        out = self.CNN_out(out)
        
        # post process
        out = torch.squeeze(out)
        if self.vmin != None and self.vmax != None:
            out = ((self.vmax-self.vmin)*torch.tanh(out) + (self.vmax+self.vmin))/2
        out = torch.squeeze(out)*self.unit
        h_v,w_v = out.shape
        h_v0,w_v0 = self.h_v0,self.w_v0
        out = out[(h_v-h_v0)//2:(h_v-h_v0)//2+h_v0,
                  (w_v-w_v0)//2:(w_v-w_v0)//2+w_v0]
        return out

# model (input with a 2D random noise)
class CNN_modify(torch.nn.Module):
    def __init__(self,model_shape,
                 in_channels  = [8,32,16],
                 vmin=None,vmax=None,
                 dropout_prob=0,
                 random_state_num = 100,
                 unit = 1000,
                 device="cpu"):
        """
            model_shape (tuple) : the shape of velocity model
            in_channels (list)  : the input and output channels for each CNN block
            vmin (float)        : the minimum velocity of output
            vmax (float)        : the maximum velocity of output
            dropout_prob (float): dropout
            device (optional)   : cpu or cuda
        """
        super(CNN_modify,self).__init__()
        self.vmin = vmin
        self.vmax = vmax
        self.device = device
        self.unit = unit
        
        # model setting
        self.layer_num  = layer_num = len(in_channels)-1
        h_in            = model_shape[0]
        w_in            = model_shape[1]
        self.h_v0       = model_shape[0]
        self.w_v0       = model_shape[1]
        
        # neural network blocks
        self.in_features = random_state_num
        
        self.FNN_in = nn.Sequential(
            nn.Linear(in_features=self.in_features,out_features=h_in*w_in*in_channels[0],bias=False),
            nn.Unflatten(0,(-1,in_channels[0],h_in,w_in)),
            nn.LeakyReLU(0.1)
        )
        
        self.CNN_Blocks = nn.ModuleList()
        for i in range(layer_num):
            self.CNN_Blocks.append(
                nn.Sequential(
                    nn.Conv2d(in_channels = in_channels[i],out_channels=in_channels[i+1],kernel_size=3,stride=1,padding="same",bias=False),
                    nn.LeakyReLU(0.1),
                    nn.Dropout(p=dropout_prob)  # add dropout layer
                )
            )
        
        self.CNN_out = nn.Sequential(
            nn.Conv2d(in_channels = in_channels[-1],out_channels=1,kernel_size=3,stride=1,padding="same",bias=False)
        )
        
        # latent variable
        torch.manual_seed(1234)
        self.random_latent_vector = torch.rand(self.in_features).to(self.device)
    
    def forward(self):
        # neural network generation
        out = self.FNN_in(self.random_latent_vector)
        for i in range(self.layer_num):
            out = self.CNN_Blocks[i](out)
        out = self.CNN_out(out)
        
        # post process
        out = torch.squeeze(out)
        if self.vmin != None and self.vmax != None:
            out = ((self.vmax-self.vmin)*torch.tanh(out) + (self.vmax+self.vmin))/2
        out = torch.squeeze(out)*self.unit
        h_v,w_v = out.shape
        h_v0,w_v0 = self.h_v0,self.w_v0
        out = out[(h_v-h_v0)//2:(h_v-h_v0)//2+h_v0,
                  (w_v-w_v0)//2:(w_v-w_v0)//2+w_v0]
        return out


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import torch
from ADFWI.utils import numpy2tensor

def thomsen_init(vp,vs,rho,eps,delta,gamma,device,dtype=torch.float32):
    """Initialize the thomsen parameters
    """
    vp      = numpy2tensor(vp,dtype).to(device)
    vs      = numpy2tensor(vs,dtype).to(device)
    rho     = numpy2tensor(rho,dtype).to(device)
    eps     = numpy2tensor(eps,dtype).to(device)
    delta   = numpy2tensor(delta,dtype).to(device)
    gamma   = numpy2tensor(gamma,dtype).to(device)
    return vp,vs,rho,eps,delta,gamma

def elastic_moduli_init(nz,nx,device,dtype=torch.float32):
    """ full anisotropic medium with 21 independent parameters
    Description:
    ------------
        *************************************************
        *   C11     C12     C13     C14     C15     C16 *
        *           C22     C23     C24     C25     C26 * 
        *                   C33     C34     C35     C36 * 
        *                           C44     C45     C46 * 
        *                                   C55     C66 * 
        *                                           C66 *
        *************************************************
    """
    # The 21 independent elastic parameters for full anisotropic models
    C11,C12,C13,C14,C15,C16 = torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device)
    C22,C23,C24,C25,C26     = torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device)
    C33,C34,C35,C36         = torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device)
    C44,C45,C46             = torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device)
    C55,C56                 = torch.zeros((nz,nx),dtype=dtype).to(device),torch.zeros((nz,nx),dtype=dtype).to(device)
    C66                     = torch.zeros((nz,nx),dtype=dtype).to(device)
    CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
    return CC

def vs_vp_to_Lame(vp,vs,rho):
    """Transform Thomsen parameter to Lame constant
    Parameters
    ----------
        vp : P-wave velocity along the respective axis
        vs : S-wave velocity along the respective axis
        rho: density
        nx : grids number along X-axis
        nz : grids number along Z-axis

    Returns
    -------
        mu  : Lame constant:shear modulus $\mu$
        lamu: Lame constant $\lambda$
        lamu: lambda - 2*mu
        b   : buoyancy = 1/density
    """
    mu   = vs**2*rho
    lamu = vp**2*rho
    lam  = lamu-2*mu
    b    = 1/rho
    return mu,lamu,lam,b


def thomsen_to_elastic_moduli(alpha_thomsen,beta_thomsen,rho,eps_thomsen,delta_thomsen,gamma_thomsen):
    """ Transform Thomsen parameter to Elastic Moduli
    Description:
    ----------
    For Transverse isotropy case, only 5 independent components: C11,C13,C33,C44,C66 ,and C12=C11-2*C66
        *********************************************************
        *   C11         C11-2C66    C13     0       0       0   *
        *   C11-2C66    C11         C13     0       0       0   *
        *   C13         C13         C33     0       0       0   *
        *   0           0           0       C44     0       0   *
        *   0           0           0       0       C44     0   *
        *   0           0           0       0       0       C66 *
        *********************************************************

    Parameters
    ----------
        alpha_thomsen: vp
        beta_thomsen : vs
        rho          : density
        eps_thomsen  : anisotropy parameter
        delta_thomsen: anisotropy parameter
        gamma_thomsen: anisotropy parameter

    Returns
    -------
        C11 : Elastic Moduli
        C13 : Elastic Moduli
        C33 : Elastic Moduli
        C44 : Elastic Moduli
        C66 : Elastic Moduli
    """
    C33 = alpha_thomsen**2*rho
    C44 =  beta_thomsen**2*rho
    C11 = C33*(1+2*eps_thomsen)
    C66 = C44*(1+2*gamma_thomsen)
    C13 = torch.sqrt(2*C33*(C33-C44)*delta_thomsen + (C33-C44)**2) - C44
    return C11,C13,C33,C44,C66

def elastic_moduli_to_thomsen(C11,C13,C33,C44,C66,rho):
    """ Transform Elastic Moduli to Thomsen parameter
    Parameters
    ----------
        C11 : Elastic Moduli
        C13 : Elastic Moduli
        C33 : Elastic Moduli
        C44 : Elastic Moduli
        C66 : Elastic Moduli
        rho : density
    
    Returns
    -------
        alpha_thomsen: vp
        beta_thomsen : vs
        eps_thomsen  : anisotropy parameter
        delta_thomsen: anisotropy parameter
        gamma_thomsen: anisotropy parameter
    """
    alpha_thomsen   = torch.sqrt(C33/rho)
    beta_thomsen    = torch.sqrt(C44/rho)
    eps_thomsen     = (C11 - C33)/(2*C33)
    gamma_thomsen   = (C66 - C44)/(2*C44)
    delta_thomsen   = ((C13+C44)**2 - (C33-C44)**2)/(2*(C33)*(C33 - C44))
    return alpha_thomsen,beta_thomsen,eps_thomsen,delta_thomsen,gamma_thomsen

def elastic_moduli_for_isotropic(CC):
    """ For  isotropy case, only 2 independent components
    Description
    ------------
        *****************************************************************
        *   lambda+2mu  lambda      lambda      0       0       0       *
        *   C11-2C66    lambda+2mu  lambda      0       0       0       *
        *   lambda      lambda      lambda+2mu  0       0       0       *
        *   0           0           0           mu      0       0       *
        *   0           0           0           0       mu      0       *
        *   0           0           0           0       0       mu      *
        *****************************************************************
    """
    [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = CC
    C22 = C11.clone()
    C55 = C44.clone()
    C23 = C13.clone()
    C12 = (C11 - 2*C66).clone()
    CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
    return CC

def elastic_moduli_for_TI(CC,anisotropic_type="VTI"):
    """ For Transverse isotropy case, only 5 independent components: C11,C13,C33,C44,C66 ,and C12=C11-2*C66
    Description
    -------------VTI
        *********************************************************
        *   C11         C11-2C66    C13     0       0       0   *
        *   C11-2C66    C11         C13     0       0       0   *
        *   C13         C13         C33     0       0       0   *
        *   0           0           0       C44     0       0   *
        *   0           0           0       0       C44     0   *
        *   0           0           0       0       0       C66 *
        *********************************************************
        
    -------------HTI 
        *********************************************************
        *   C33         C13         C13      0       0       0   *
        *   C13         C11         C12      0       0       0   *
        *   C13         C12         C33      0       0       0   *
        *   0           0           0        C66     0       0   *
        *   0           0           0        0       C55     0   *
        *   0           0           0        0       0       C55 *
        *********************************************************
    """
    [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = CC
    C22 = C11.clone()
    C55 = C44.clone()
    C23 = C13.clone()
    C12 = (C11 - 2*C66).clone()
    # HTI Rotated by VTI anticlockwise (Y) pi/2
    if anisotropic_type.lower() in ["vti"]:
        CC = [C11,C12,C13,C14,C15,C16,
                  C22,C23,C24,C25,C26,
                      C33,C34,C35,C36,
                          C44,C45,C46,
                              C55,C56,
                                  C66]
        # [C11,C12,C13,C15,C22,C23,C33,C35,C44,C55,C66]
    elif anisotropic_type.lower() in ["hti"]:
        CC = [C33,C13,C13,C14,C15,C16,
                  C11,C12,C24,C25,C26,
                      C11,C34,C35,C36,
                          C66,C45,C46,
                              C55,C56,
                                  C55]
        # [C33,C13,C13,C15,C11,C12,C11,C35,C66,C55,C55]
    return CC


def parameter_staggered_grid(mu,b,C44,C55,C66,nx,nz):
    """ Staggered mesh settings for model parameters
    Parameters
    ----------
        mu  : Lame constant:shear modulus $\mu$
        b   : buoyancy = 1/density 
    Returns
    -------
        bx  : staggered grid for b  in X-axis
        bz  : staggered grid for b  in Z-axis
        muxz: staggered grid for mu
        C44 : staggered grid for C44
        C55 : staggered grid for C55
        C66 : staggered grid for C66
    """
    bx   = 0.5*(b[:,0:nx-1]+b[:,1:nx])
    bz   = 0.5*(b[0:nz-1,:]+b[1:nz,:])

    muxz = 0.2*(mu[1:nz-1,1:nx-1]  + mu[2:nz,1:nx-1] + mu[1:nz-1,2:nx]      +\
                mu[2:nz  ,1:nx-1]  + mu[2:nz,2:nx])

    C44  =  0.2*(C44[1:nz-1,1:nx-1] + C44[2:nz,1:nx-1] + C44[1:nz-1,2:nx]    +\
                C44[2:nz  ,1:nx-1] + C44[2:nz,2:nx])

    C55  =  0.2*(C55[1:nz-1,1:nx-1] + C55[2:nz,1:nx-1] + C55[1:nz-1,2:nx]    +\
                C55[2:nz  ,1:nx-1] + C55[2:nz,2:nx])

    C66  =  0.2*(C66[1:nz-1,1:nx-1] + C66[2:nz,1:nx-1] + C66[1:nz-1,2:nx]    +\
                C66[2:nz  ,1:nx-1] + C66[2:nz,2:nx])
    return bx,bz,muxz,C44,C55,C66



from .base import Regularization
from .tikhonov_1order import Tikhonov_1order as regularization_Tikhonov_1order
from .tikhonov_2order import Tikhonov_2order as regularization_Tikhonov_2order
from .tv_1order import TV_1order             as regularization_TV_1order
from .tv_2order import TV_2order             as regularization_TV_2order


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-19 17:05:45
* LastEditors: LiuFeng
* LastEditTime: 2024-05-07 16:05:36
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import matplotlib.pyplot as plt
import numpy as np
import os 
from ADFWI.utils.utils import gpu2cpu


def plot_vp_vs_rho(vp,vs,rho,dx=-1,dz=-1,figsize=(14,4),wspace=0.2,hspace=0.2,
                   cmap='jet',title="",
                   vp_min   = None  ,vp_max =None,
                   vs_min   = None  ,vs_max =None,
                   rho_min  = None  ,rho_max=None,
                   tick_param  = {'labelsize':15},label_param = {'family':'Times New Roman','weight':'normal','size': 15},
                   colorbar_param = {'labelsize':12},title_param = {'family':'Times New Roman','weight':'normal','size': 20},
                   cbar_pad_fraction=0.1,cbar_height=0.05,
                   show=True,save_path="",save_dpi=300):
    """plot vp/vs and density
        plot a data section using matplotlib.pcolormesh or imshow
    
    Parameters:
    ----------------
        - vp (ndarrary or Tensor)           : p-wave velocity (m/s)
        - vs (ndarrary or Tensor)           : s-wave velocity (m/s)
        - rho (ndarrary or Tensor)          : densicyt (kg/m^3)
        - dx (float,optional)               : the spatial sampling interval along x-axis. Default: -1
        - dz (float,optional)               : the spatial sampling interval along z-axis. Default: -1
        - figsize (tuple,optional)          : the size of figure. Default (14,4)
        - wspace (float,optional)           : the width of the padding between subplots, as a fraction of the average Axes width.   Default: 0.2
        - hspace (float,optional)           : the height of the padding between subplots, as a fraction of the average Axes height. Default: 0.2
        - cmap (str or Colormap,optional)   : the Colormap instance or registered colormap name used to map scalar data to colors.  Default: 'jet'
        - tick_param (dict,optional)        : the fontdict for ticks
        - label_param (dict,optional)       : the fontdict for label
        - colorbar_param (dict,optional)    : the fontdict for colorbar
        - title_param (dict,optional)       : the fontdict for title
        - cbar_pad_fraction (float,optional): the padding interval for colorbar and main figure
        - cbar_height (float,optional)      : the height of the colorbar
        - show (bool,optional)              : showing the figure or not. Default True
        - save_path (str,optional)          : the saving path for the figure. Defualt:""
        - save_dpi (int,optional)           : the saving resolution for the figure. Default:300 dpi
    Retures
    -----------------
    None
    """
    vp = gpu2cpu(vp)
    vs = gpu2cpu(vs)
    rho = gpu2cpu(rho)
    
    nz,nx = vp.shape
    x = np.arange(nx)*dx/1000
    z = np.arange(nz)*dz/1000
    X,Z = np.meshgrid(x,z)
    
    ######################################################
    fig,ax = plt.subplots(1,3,figsize=figsize)
    if dx>0 and dz >0:
        if vp_min is not None and vp_max is not None:
            im1 = ax[0].pcolormesh(X,Z,vp,cmap=cmap,vmin=vp_min,vmax=vp_max)
        else:
            im1 = ax[0].pcolormesh(X,Z,vp,cmap=cmap)
        ax[0].invert_yaxis()
        ax[0].set_xlabel("X (km)",**label_param)
        ax[0].set_ylabel("Z (km)",**label_param)
    else:
        if vp_min is not None and vp_max is not None:
            im1 = ax[0].imshow(vp,cmap=cmap,vmin=vp_min,vmax=vp_max)
        else:
            im1 = ax[0].imshow(vp,cmap=cmap)
        ax[0].set_xlabel("X",**label_param)
        ax[0].set_ylabel("Z",**label_param)
    ax[0].set_title("vp (m/s)",**title_param)
    ax[0].tick_params(**tick_param)
    ax[0].axis("scaled")
    cax1 = fig.add_axes([ax[0].get_position().x0,ax[0].get_position().y0-cbar_pad_fraction,
                         ax[0].get_position().width,cbar_height])
    cbar1 = plt.colorbar(im1,cax=cax1,orientation='horizontal')
    cbar1.ax.tick_params(**colorbar_param)

    ######################################################
    if dx>0 and dz >0:
        if vs_min is not None and vs_max is not None:
            im2 = ax[1].pcolormesh(X,Z,vs,cmap=cmap,vmin=vs_min,vmax=vs_max)
        else:
            im2 = ax[1].pcolormesh(X,Z,vs,cmap=cmap)
        ax[1].invert_yaxis()
        ax[1].set_xlabel("X (km)",**label_param)
        ax[1].set_ylabel("Z (km)",**label_param)
    else:
        if vs_min is not None and vs_max is not None:
            im2 = ax[1].imshow(vs,cmap=cmap,vmin=vs_min,vmax=vs_max)
        else:
            im2 = ax[1].imshow(vs,cmap=cmap)
        ax[1].set_xlabel("X",**label_param)
        ax[1].set_ylabel("Z",**label_param)
    ax[1].set_title("vs (m/s)",**title_param)
    ax[1].tick_params(**tick_param)
    ax[1].axis("scaled")
    cax2 = fig.add_axes([ax[1].get_position().x0,ax[1].get_position().y0-cbar_pad_fraction,
                         ax[1].get_position().width,cbar_height])
    cbar2 = plt.colorbar(im2,cax=cax2,orientation='horizontal')
    cbar2.ax.tick_params(**colorbar_param)
    
    ######################################################
    if dx>0 and dz >0:
        if rho_min is not None and rho_max is not None:
            im3 = ax[2].pcolormesh(X,Z,rho,cmap=cmap,vmin=rho_min,vmax=rho_max)
        else:
            im3 = ax[2].pcolormesh(X,Z,rho,cmap=cmap)
        ax[2].invert_yaxis()
        ax[2].set_xlabel("X (km)",**label_param)
        ax[2].set_ylabel("Z (km)",**label_param)
    else:
        if rho_min is not None and rho_max is not None:
            im3 = ax[2].imshow(rho,cmap=cmap,vmin=rho_min,vmax=rho_max)
        else:
            im3 = ax[2].imshow(rho,cmap=cmap)
        ax[2].set_xlabel("X (km)",**label_param)
        ax[2].set_ylabel("Z (km)",**label_param)

    ax[2].set_title(r"$\rho$ (kg/m^3)",**title_param)
    ax[2].tick_params(**tick_param)
    ax[2].axis("scaled")
    cax3 = fig.add_axes([ax[2].get_position().x0,ax[2].get_position().y0-cbar_pad_fraction,
                         ax[2].get_position().width,cbar_height])
    cbar3 = plt.colorbar(im3,cax=cax3,orientation='horizontal')
    cbar3.ax.tick_params(**colorbar_param)
    
    if not title == "":
        fig.suptitle(title,**title_param)
        
    plt.subplots_adjust(hspace=hspace,wspace=wspace)
    if not save_path == "":
        plt.savefig(save_path,dpi=save_dpi,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()

def plot_vp_rho(vp,rho,dx=-1,dz=-1,figsize=(10,5),wspace=0.2,hspace=0.2,
                   cmap='jet',
                   tick_param       = {'labelsize':15},label_param = {'family':'Times New Roman','weight':'normal','size': 15},
                   colorbar_param   = {'labelsize':15},title_param = {'family':'Times New Roman','weight':'normal','size': 20},
                   cbar_pad_fraction=0.1,cbar_height=0.05,
                   show=True,save_path="",save_dpi=300):
    """plot vp and rho
    plot a data section using matplotlib.pcolormesh or imshow
    
    Parameters:
    ----------------
        - vp (ndarrary or Tensor)           : p-wave velocity (m/s)
        - rho (ndarrary or Tensor)          : densicyt (kg/m^3)
        - dx (float,optional)               : the spatial sampling interval along x-axis. Default: -1
        - dz (float,optional)               : the spatial sampling interval along z-axis. Default: -1
        - figsize (tuple,optional)          : the size of figure. Default (14,4)
        - wspace (float,optional)           : the width of the padding between subplots, as a fraction of the average Axes width.   Default: 0.2
        - hspace (float,optional)           : the height of the padding between subplots, as a fraction of the average Axes height. Default: 0.2
        - cmap (str or Colormap,optional)   : the Colormap instance or registered colormap name used to map scalar data to colors.  Default: 'jet'
        - tick_param (dict,optional)        : the fontdict for ticks
        - label_param (dict,optional)       : the fontdict for label
        - colorbar_param (dict,optional)    : the fontdict for colorbar
        - title_param (dict,optional)       : the fontdict for title
        - cbar_pad_fraction (float,optional): the padding interval for colorbar and main figure
        - cbar_height (float,optional)      : the height of the colorbar
        - show (bool,optional)              : showing the figure or not. Default True
        - save_path (str,optional)          : the saving path for the figure. Defualt:""
        - save_dpi (int,optional)           : the saving resolution for the figure. Default:300 dpi
    Retures
    -----------------
    None
    """
    vp          = gpu2cpu(vp)
    rho         = gpu2cpu(rho)
    
    nz,nx = vp.shape
    x = np.arange(nx)*dx/1000
    z = np.arange(nz)*dz/1000
    X,Z = np.meshgrid(x,z)
    
    ######################################################
    fig,ax = plt.subplots(1,2,figsize=figsize)
    if dx>0 and dz >0:
        im1 = ax[0].pcolormesh(X,Z,vp,cmap=cmap)
        ax[0].invert_yaxis()
        ax[0].set_xlabel("X (km)",**label_param)
        ax[0].set_ylabel("Z (km)",**label_param)
    else:
        im1 = ax[0].imshow(vp)
        ax[0].set_xlabel("X",**label_param)
        ax[0].set_ylabel("Z",**label_param)
    cax1 = fig.add_axes([ax[0].get_position().x0,ax[0].get_position().y0-cbar_pad_fraction,
                         ax[0].get_position().width,cbar_height])
    cbar1 = plt.colorbar(im1,cax=cax1,orientation='horizontal')
    cbar1.ax.tick_params(**colorbar_param)
    ax[0].set_title("vp (m/s)",**title_param)
    ax[0].tick_params(**tick_param)
    ax[0].axis("scaled")
    
    ######################################################
    if dx>0 and dz >0:
        im2 = ax[1].pcolormesh(X,Z,rho,cmap=cmap)
        ax[1].invert_yaxis()
        ax[1].set_xlabel("X (km)",**label_param)
        ax[1].set_ylabel("Z (km)",**label_param)
    else:
        im2 = ax[1].imshow(rho,cmap=cmap)
        ax[1].set_xlabel("X",**label_param)
        ax[1].set_ylabel("Z",**label_param)
    cax2 = fig.add_axes([ax[1].get_position().x0,ax[1].get_position().y0-cbar_pad_fraction,
                         ax[1].get_position().width,cbar_height])
    cbar2 = plt.colorbar(im2,cax=cax2,orientation='horizontal')
    cbar2.ax.tick_params(**colorbar_param)
    ax[1].set_title(r"$\rho$ (kg/m^3)",**title_param)
    ax[1].tick_params(**tick_param)
    ax[1].axis("scaled")
    
    
    plt.subplots_adjust(hspace=hspace,wspace=wspace)
    if not save_path == "":
        plt.savefig(save_path,dpi=save_dpi,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()


def plot_eps_delta_gamma(eps,delta,gamma,dx=-1,dz=-1,figsize=(14,4),wspace=0.2,hspace=0.2,
                   cmap='jet',
                   tick_param  = {'labelsize':15},label_param = {'family':'Times New Roman','weight':'normal','size': 15},
                   colorbar_param = {'labelsize':15},title_param = {'family':'Times New Roman','weight':'normal','size': 20},
                   cbar_pad_fraction=0.1,cbar_height=0.05,
                   show=True,save_path="",save_dpi=300):
    """plot epsilon/delta and gamma
    plot a data section using matplotlib.pcolormesh or imshow
    
    Parameters:
    ----------------
        - eps   (ndarrary or Tensor)        : the anisotropic parameter
        - delta (ndarrary or Tensor)        : the anisotropic parameter
        - gamma (ndarrary or Tensor)        : the anisotropic parameter
        - dx (float,optional)               : the spatial sampling interval along x-axis. Default: -1
        - dz (float,optional)               : the spatial sampling interval along z-axis. Default: -1
        - figsize (tuple,optional)          : the size of figure. Default (14,4)
        - wspace (float,optional)           : the width of the padding between subplots, as a fraction of the average Axes width.   Default: 0.2
        - hspace (float,optional)           : the height of the padding between subplots, as a fraction of the average Axes height. Default: 0.2
        - cmap (str or Colormap,optional)   : the Colormap instance or registered colormap name used to map scalar data to colors.  Default: 'jet'
        - tick_param (dict,optional)        : the fontdict for ticks
        - label_param (dict,optional)       : the fontdict for label
        - colorbar_param (dict,optional)    : the fontdict for colorbar
        - title_param (dict,optional)       : the fontdict for title
        - cbar_pad_fraction (float,optional): the padding interval for colorbar and main figure
        - cbar_height (float,optional)      : the height of the colorbar
        - show (bool,optional)              : showing the figure or not. Default True
        - save_path (str,optional)          : the saving path for the figure. Defualt:""
        - save_dpi (int,optional)           : the saving resolution for the figure. Default:300 dpi
    Retures
    -----------------
    None
    """
    eps     = gpu2cpu(eps)
    delta   = gpu2cpu(delta)
    gamma   = gpu2cpu(gamma)
    
    nz,nx = eps.shape
    x = np.arange(nx)*dx/1000
    z = np.arange(nz)*dz/1000
    X,Z = np.meshgrid(x,z)
    
    ######################################################
    fig,ax = plt.subplots(1,3,figsize=figsize)
    if dx>0 and dz >0:
        im1 = ax[0].pcolormesh(X,Z,eps,cmap=cmap)
        ax[0].invert_yaxis()
        ax[0].set_xlabel("X (km)",**label_param)
        ax[0].set_ylabel("Z (km)",**label_param)
    else:
        im1 = ax[0].imshow(eps,cmap=cmap)
        ax[0].set_xlabel("X",**label_param)
        ax[0].set_ylabel("Z",**label_param)
    cax1 = fig.add_axes([ax[0].get_position().x0,ax[0].get_position().y0-cbar_pad_fraction,
                         ax[0].get_position().width,cbar_height])
    cbar1 = plt.colorbar(im1,cax=cax1,orientation='horizontal')
    cbar1.ax.tick_params(**colorbar_param)
    ax[0].set_title("eps",**title_param)
    ax[0].tick_params(**tick_param)
    ax[0].axis("scaled")
    
    ######################################################
    if dx>0 and dz >0:
        im2 = ax[1].pcolormesh(X,Z,delta,cmap=cmap)
        ax[1].invert_yaxis()
        ax[1].set_xlabel("X (km)",**label_param)
        ax[1].set_ylabel("Z (km)",**label_param)
    else:
        im2 = ax[0].imshow(delta,cmap=cmap)
        ax[1].set_xlabel("X",**label_param)
        ax[1].set_ylabel("Z",**label_param)
    cax2 = fig.add_axes([ax[1].get_position().x0,ax[1].get_position().y0-cbar_pad_fraction,
                         ax[1].get_position().width,cbar_height])
    cbar2 = plt.colorbar(im2,cax=cax2,orientation='horizontal')
    cbar2.ax.tick_params(**colorbar_param)
    ax[1].set_title("delta",**title_param)
    ax[1].tick_params(**tick_param)
    ax[1].axis("scaled")
    
    ######################################################
    if dx>0 and dz >0:
        im3 = ax[2].pcolormesh(X,Z,gamma,cmap=cmap)
        ax[2].invert_yaxis()
        ax[2].set_xlabel("X (km)",**label_param)
        ax[2].set_ylabel("Z (km)",**label_param)
    else:
        im3 = ax[2].imshow(gamma,cmap=cmap)
        ax[2].set_xlabel("X",**label_param)
        ax[2].set_ylabel("Z",**label_param)
    cax3 = fig.add_axes([ax[2].get_position().x0,ax[2].get_position().y0-cbar_pad_fraction,
                         ax[2].get_position().width,cbar_height])
    cbar3 = plt.colorbar(im3,cax=cax3,orientation='horizontal')
    cbar3.ax.tick_params(**colorbar_param)
    ax[2].set_title("gamma",**title_param)
    ax[2].tick_params(**tick_param)
    ax[2].axis("scaled")
    
    plt.subplots_adjust(hspace=hspace,wspace=wspace)
    if not save_path == "":
        plt.savefig(save_path,dpi=save_dpi,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()


def plot_lam_mu(lam,mu,dx=-1,dz=-1,figsize=(10,5),wspace=0.2,hspace=0.2,
                   cmap='jet',
                   tick_param       = {'labelsize':15},label_param = {'family':'Times New Roman','weight':'normal','size': 15},
                   colorbar_param   = {'labelsize':15},title_param = {'family':'Times New Roman','weight':'normal','size': 20},
                   cbar_pad_fraction=0.1,cbar_height=0.05,
                   show=True,save_path="",save_dpi=300):
    """plot lambda and mu
    plot a data section using matplotlib.pcolormesh or imshow
    
    Parameters:
    ----------------
        - lam (ndarrary or Tensor)          : the lame constant parameter
        - mu (ndarrary or Tensor)           : the lame constant parameter
        - dx (float,optional)               : the spatial sampling interval along x-axis. Default: -1
        - dz (float,optional)               : the spatial sampling interval along z-axis. Default: -1
        - figsize (tuple,optional)          : the size of figure. Default (14,4)
        - wspace (float,optional)           : the width of the padding between subplots, as a fraction of the average Axes width.   Default: 0.2
        - hspace (float,optional)           : the height of the padding between subplots, as a fraction of the average Axes height. Default: 0.2
        - cmap (str or Colormap,optional)   : the Colormap instance or registered colormap name used to map scalar data to colors.  Default: 'jet'
        - tick_param (dict,optional)        : the fontdict for ticks
        - label_param (dict,optional)       : the fontdict for label
        - colorbar_param (dict,optional)    : the fontdict for colorbar
        - title_param (dict,optional)       : the fontdict for title
        - cbar_pad_fraction (float,optional): the padding interval for colorbar and main figure
        - cbar_height (float,optional)      : the height of the colorbar
        - show (bool,optional)              : showing the figure or not. Default True
        - save_path (str,optional)          : the saving path for the figure. Defualt:""
        - save_dpi (int,optional)           : the saving resolution for the figure. Default:300 dpi
    Retures
    -----------------
    None
    """
    lam     = gpu2cpu(lam)
    mu      = gpu2cpu(mu)
    
    nz,nx = lam.shape
    x = np.arange(nx)*dx/1000
    z = np.arange(nz)*dz/1000
    X,Z = np.meshgrid(x,z)
    
    ######################################################
    fig,ax = plt.subplots(1,2,figsize=figsize)
    if dx>0 and dz >0:
        im1 = ax[0].pcolormesh(X,Z,lam,cmap=cmap)
        ax[0].invert_yaxis()
        ax[0].set_xlabel("X (km)",**label_param)
        ax[0].set_ylabel("Z (km)",**label_param)
    else:
        im1 = ax[0].imshow(lam)
        ax[0].set_xlabel("X",**label_param)
        ax[0].set_ylabel("Z",**label_param)
    cax1 = fig.add_axes([ax[0].get_position().x0,ax[0].get_position().y0-cbar_pad_fraction,
                         ax[0].get_position().width,cbar_height])
    cbar1 = plt.colorbar(im1,cax=cax1,orientation='horizontal')
    cbar1.ax.tick_params(**colorbar_param)
    ax[0].set_title(r"$\lambda$",**title_param)
    ax[0].tick_params(**tick_param)
    ax[0].axis("scaled")
    
    ######################################################
    if dx>0 and dz >0:
        im2 = ax[1].pcolormesh(X,Z,mu,cmap=cmap)
        ax[1].invert_yaxis()
        ax[1].set_xlabel("X (km)",**label_param)
        ax[1].set_ylabel("Z (km)",**label_param)
    else:
        im2 = ax[1].imshow(mu,cmap=cmap)
        ax[1].set_xlabel("X",**label_param)
        ax[1].set_ylabel("Z",**label_param)
    cax2 = fig.add_axes([ax[1].get_position().x0,ax[1].get_position().y0-cbar_pad_fraction,
                         ax[1].get_position().width,cbar_height])
    cbar2 = plt.colorbar(im2,cax=cax2,orientation='horizontal')
    cbar2.ax.tick_params(**colorbar_param)
    ax[1].set_title(r"$\mu$",**title_param)
    ax[1].tick_params(**tick_param)
    ax[1].axis("scaled")
    
    
    plt.subplots_adjust(hspace=hspace,wspace=wspace)
    if not save_path == "":
        plt.savefig(save_path,dpi=save_dpi,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()
        
def plot_model(vel_model,title,dx=-1,dz=-1,
                figsize=(8,8),wspace=0.2,hspace=0.2,
                cmap='jet',
                tick_param       = {'labelsize':15},label_param = {'family':'Times New Roman','weight':'normal','size': 15},
                colorbar_param   = {'labelsize':15},title_param = {'family':'Times New Roman','weight':'normal','size': 20},
                cbar_pad_fraction=0.12,cbar_height=0.03,
                vmin = None,vmax=None,
                show=True,save_path="",save_dpi=300):
    
    vel_model     = gpu2cpu(vel_model)
    nz,nx = vel_model.shape
    x = np.arange(nx)*dx/1000
    z = np.arange(nz)*dz/1000
    X,Z = np.meshgrid(x,z)
    
    ######################################################
    fig,ax = plt.subplots(1,1,figsize=figsize)
    if dx>0 and dz >0:
        if vmin is not None and vmax is not None:
            im1 = ax.pcolormesh(X,Z,vel_model,cmap=cmap,vmin=vmin,vmax=vmax)
        else:
            im1 = ax.pcolormesh(X,Z,vel_model,cmap=cmap)
        ax.invert_yaxis()
        ax.set_xlabel("X (km)",**label_param)
        ax.set_ylabel("Z (km)",**label_param)
    else:
        if vmin is not None and vmax is not None:
            im1 = ax.imshow(vel_model,cmap=cmap,vmin=vmin,vmax=vmax)
        else:
            im1 = ax.imshow(vel_model,cmap=cmap)
        ax.set_xlabel("X",**label_param)
        ax.set_ylabel("Z",**label_param)
    
    ax.set_title(title,**title_param)
    ax.tick_params(**tick_param)
    ax.axis("scaled")
    
    cax1 = fig.add_axes([ax.get_position().x0,ax.get_position().y0-cbar_pad_fraction,
                         ax.get_position().width,cbar_height])
    cbar1 = plt.colorbar(im1,cax=cax1,orientation='horizontal')
    cbar1.ax.tick_params(**colorbar_param)
    
    plt.subplots_adjust(hspace=hspace,wspace=wspace)
    if not save_path == "":
        plt.savefig(save_path,dpi=save_dpi,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()
    return


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import torch
from torch.autograd import Function

"""
    NIM misfit
        https://github.com/YangFangShu/FWIGAN/blob/main/FWIGAN/utils/misfit.py
"""

def transform(f, g, trans_type, theta):
    """
    Apply a transformation to make f and g non-negative.
    
    Args:
        f, g (Tensor): Seismic data, shape [num_time_steps, num_shots*num_receivers_per_shot].
        trans_type (str): Transformation type ('linear', 'abs', 'square', 'exp', 'softplus').
        theta (float): Scalar parameter for transformation.
        
    Returns:
        mu (Tensor): Transformed f.
        nu (Tensor): Transformed g.
        d (Tensor): Derivative of transformed f for potential use in gradient-based methods.
    """
    c = 0.0  # Initialize the offset constant
    device = f.device
    if trans_type == 'linear':
        # Linear transformation with offset
        min_value = torch.min(f.detach().min(), g.detach().min())
        mu, nu = f, g
        c = -min_value if min_value < 0 else 0
        c = c * theta  # Scale by theta for flexibility
        d = torch.ones(f.shape).to(device)
    elif trans_type == 'abs':
        # Absolute value transformation
        mu, nu = torch.abs(f), torch.abs(g)
        d = torch.sign(f).to(device)
    elif trans_type == 'square':
        # Squaring transformation
        mu = f * f
        nu = g * g
        d = 2 * f
    elif trans_type == 'exp':
        # Exponential transformation, scaled by theta
        mu = torch.exp(theta * f)
        nu = torch.exp(theta * g)
        d = theta * mu
    elif trans_type == 'softplus':
        # Softplus transformation for smooth non-negativity
        mu = torch.log(torch.exp(theta * f) + 1)
        nu = torch.log(torch.exp(theta * g) + 1)
        d = theta / (torch.exp(-theta * f) + 1e-18)  # Avoid division by zero
    else:
        mu, nu = f, g
        d = torch.ones(f.shape).to(device)
    # Ensure positive values for mu and nu by adding a small constant
    mu = mu + c + 1e-18
    nu = nu + c + 1e-18
    return mu, nu, d

def trace_sum_normalize(x, time_dim=0):
    """
    Normalize each trace by its sum along the specified dimension.
    
    Args:
        x (Tensor): Input tensor.
        time_dim (int): Dimension for time steps.
        
    Returns:
        Tensor: Normalized tensor.
    """
    x = x / (x.sum(dim=time_dim, keepdim=True) + 1e-18)  # Avoid division by zero
    return x

def trace_max_normalize(x, time_dim=0):
    """
    normalization with the maximum value of each trace (the value of each trace is in [-1,1] after the processing)
    note that the channel should be 1
    """
    x_max,_ = torch.max(x.abs(), dim=time_dim, keepdim=True)
    x = x / (x_max+1e-18)
    return x

class Misfit_NIM(Function):
    """
    Normalized Integration Method (NIM), computes misfit between cumulative distributions of transformed signals.
    
    Parameters:
        p (int): Norm degree, default is 2.
        trans_type (str): Type of non-negative transform, default is 'linear'.
        theta (float): Parameter for non-negative transform, default is 1.
        dt (float): Time sampling interval.
    
    Note:
        NIM is equivalent to the Wasserstein-1 distance when p=1.
    """
    def __init__(self, p=1,trans_type='linear',theta=1,dt=1):
        self.p = p
        self.trans_type = trans_type
        self.theta = theta
        self.dt = dt
    
    @staticmethod
    def forward(ctx, syn, obs, p=2, trans_type='linear', theta=1.):
        assert p >= 1, "Norm degree must be >= 1"
        assert syn.shape == obs.shape, "Shape mismatch between synthetic and observed data"
        
        # Flatten the input tensors for transformation (shape: [num_shots, num_time_steps * num_receivers])
        num_shots, num_time_steps, num_receivers = syn.shape
        
        # [num_shots, num_time_steps, num_receivers] --> [num_shots, num_receivers, num_time_steps]
        syn_transposed = syn.permute(1, 0, 2)
        obs_transposed = obs.permute(1, 0, 2) 

        # Reshape input tensors for transformation
        syn_flat = syn_transposed.reshape(num_time_steps, num_shots * num_receivers)
        obs_flat = obs_transposed.reshape(num_time_steps, num_shots * num_receivers)
        
        # Transform signals to ensure non-negativity
        mu, nu, d = transform(syn_flat, obs_flat, trans_type, theta)
        
        # Normalize each trace by its sum
        mu = trace_sum_normalize(mu, time_dim=0)
        nu = trace_sum_normalize(nu, time_dim=0)
        
        # Compute cumulative sums over the time dimension
        F = torch.cumsum(mu, dim=0)  # Keep the cumulative sum along the flattened time dimension
        G = torch.cumsum(nu, dim=0)  # Keep the cumulative sum along the flattened time dimension
        
        # Save the necessary tensors for backward computation
        ctx.save_for_backward(F - G, mu,  d)
        ctx.p = p
        ctx.num_shots = num_shots  # Save as an attribute
        ctx.num_time_steps = num_time_steps  # Save as an attribute
        ctx.num_receivers = num_receivers  # Save as an attribute
        
        return (torch.abs(F - G) ** p).sum()
    
    @staticmethod
    def backward(ctx, grad_output):
        residual, mu, d = ctx.saved_tensors
        p = ctx.p
        num_shots = ctx.num_shots
        num_time_steps = ctx.num_time_steps
        num_receivers = ctx.num_receivers
        
        if p == 1:  # Check if p is 1
            df = torch.sign(residual) * mu * d
        else:
            df = (residual) ** (p - 1) * mu * d     
        
        df = df.reshape(num_time_steps, num_shots, num_receivers).permute(1, 0, 2)
        return -df, None, None, None, None




# The following gradient are not assistant with Automatic Differentiation
# class Misfit_NIM(torch.autograd.Function):
#     """
#         Normalized Integration Method, Liu et al., 2012: the objective function measures the misfit between the integral of the absolute value, or of the square, or of the envelope of the signal.
#             F_i = \frac{\sum_{j=1}^i P(f_j)}{\sum_{j=1}^n P(f_j)}, 
#             G_i = \frac{\sum_{j=1}^i P(g_j)}{\sum_{j=1}^n P(g_j)}, 
#             \ell(f, g) = \sum_{i=1}^n |F_i - G_i|^p, 
#         where function :`P` is choosed to make the vector nonnegative, 
#             e.g. :`|x|`, `|x|^2`.
#         parameters
#         -----------
#             p (int)         : the norm degree. Default: 2 
#             trans_type (str): the nonnegative transform. Default: 'linear'
#             theta (int)     : the parameter used in nonnegtive transform. Default: 1
#         Note:
#             NIM is equivalent to Wasserstein-1 distance (Earth Mover's distance) when p = 1
#     """
#     def __init__(self,p=2,trans_type='linear',theta=1.):
#         self.p = p
#         self.trans_type = trans_type
#         self.theta = theta
    
#     @staticmethod
#     def forward(ctx, syn:torch.Tensor, obs:torch.Tensor, p=2, trans_type='linear', theta=1.):
#         assert p >= 1
#         assert syn.shape == obs.shape
#         device = syn.device
        
#         # transform for the signal syn and obs
#         p = torch.tensor(p).to(device)
#         mu, nu, d = transform(syn, obs, trans_type, theta)
        
#         # normalization with the summation of each trace
#         mu = trace_sum_normalize(mu,time_dim=1)
#         nu = trace_sum_normalize(nu,time_dim=1)
        
#         F = torch.cumsum(mu, dim=1)
#         G = torch.cumsum(nu, dim=1)
        
#         ctx.save_for_backward(F-G, mu, p, d)
        
#         return (torch.abs(F - G) ** p).sum()

#     @staticmethod
#     def backward(ctx, grad_output):
#         residual, mu, p, d = ctx.saved_tensors
#         if p == 1:
#             df = torch.sign(residual) * mu *d
#         else:
#             df = (residual) ** (p - 1) * mu * d
#         return df, None, None, None, None



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-14 09:08:12
* LastEditors: LiuFeng
* LastEditTime: 2024-05-26 10:41:48
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

import torch
import torch.nn as nn
import torch.nn.functional as F
from FyeldGenerator import generate_field
import numpy as np

# Helper that generates power-law power spectrum
def Pkgen(n):
    def Pk(k):
        return np.power(k, -n)
    return Pk

def distrib(shape):
    a = np.random.normal(loc=0, scale=1, size=shape)
    b = np.random.normal(loc=0, scale=1, size=shape)
    return a + 1j * b

def generate_grf(shape,alpha,unit_length=10,device='cpu'):
    field = generate_field(distrib, Pkgen(alpha), shape, unit_length=unit_length)
    return torch.tensor(field,dtype=torch.float32,device=device)

# Double convolution block
class DoubleConv(nn.Module):
    def __init__(self, in_channels, out_channels, mid_channels=None):
        super(DoubleConv, self).__init__()
        if not mid_channels:
            mid_channels = out_channels
            
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            # nn.LeakyReLU(0.1),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            # nn.LeakyReLU(0.1)
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)

# Downscaling block
class Down(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(Down, self).__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels),
        )

    def forward(self, x):
        return self.maxpool_conv(x)

# Upscaling block
class Up(nn.Module):
    def __init__(self, in_channels, out_channels, bilinear=True):
        super(Up, self).__init__()
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        diffY = x2.size()[2] - x1.size()[2] # CHW
        diffX = x2.size()[3] - x1.size()[3]
        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)

# Output layer
class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)

# General UNet class for any number of layers
class UNet(nn.Module):
    def __init__(self, model_shape, 
                 n_layers, 
                 base_channel, 
                 vmin=None, 
                 vmax=None, 
                 in_channels=1, out_channels=1, 
                 bilinear=False,
                 grf_initialize=False,
                 grf_alpha = 0, # power-law power spectrum
                 unit = 1000, 
                 device="cpu"
                 ):
        super(UNet, self).__init__()
        self.in_channels    = in_channels
        self.out_channels   = out_channels
        self.model_shape    = model_shape
        self.bilinear       = bilinear
        self.unit           = unit
        
        self.grf_initialize = grf_initialize
        self.grf_alpha      = grf_alpha

        self.inc    = DoubleConv(in_channels, base_channel)
        self.downs  = nn.ModuleList()
        self.ups    = nn.ModuleList()

        # Create down blocks
        channels = base_channel
        for _ in range(n_layers):
            self.downs.append(Down(channels, channels * 2))
            channels *= 2

        # Create up blocks
        factor = 2 if bilinear else 1
        for _ in range(n_layers):
            self.ups.append(Up(channels, channels // 2 // factor, bilinear))
            channels //= 2

        self.outc = OutConv(channels, out_channels)

        # Random latent variable for input
        self.device = device
        self.vmin = vmin
        self.vmax = vmax
        self.h0, self.w0 = model_shape
        if grf_initialize:
            self.random_latent_vector = self._grf_initialize()
        else:
            self.random_latent_vector = self._random_initialize()
    
    def _random_initialize(self):
        torch.manual_seed(1234)
        return torch.rand(1, 1, self.h0, self.w0).to(self.device)
    
    def _grf_initialize(self):
        return generate_grf(self.model_shape,self.grf_alpha,device=self.device).unsqueeze(0).unsqueeze(0)

    def forward(self,x=None):
        if x is None:
            x = self.random_latent_vector
        x1 = self.inc(x)
        # down-sampling
        downs_outputs = [x1]
        for down in self.downs:
            downs_outputs.append(down(downs_outputs[-1]))
        # up-sampling
        x = downs_outputs[-1]
        for i in range(len(self.ups)):
            x = self.ups[i](x, downs_outputs[-2-i])
        # output velocity model
        out = self.outc(x)
        
        # post process
        out = torch.squeeze(out)
        if self.vmin is not None and self.vmax is not None:
            out = ((self.vmax - self.vmin) * torch.tanh(out) + (self.vmax + self.vmin)) / 2
        out = torch.squeeze(out) * self.unit
        return out


# def approximate_kv(v, x):
#     """
#     Approximate the modified Bessel function of the second kind (Kv) for PyTorch.
#     Args:
#         v (torch.Tensor): Order of the Bessel function (v > 0).
#         x (torch.Tensor): Input value (x > 0).
#     Returns:
#         torch.Tensor: Approximated Kv(v, x).
#     """
#     # Avoid division by zero
#     x = torch.clamp(x, min=1e-10)

#     # Asymptotic expansion for large x
#     large_x = x > 10  # Threshold for large x
#     kv_large = torch.exp(-x) * torch.sqrt(torch.pi / (2 * x))

#     # Small x approximation
#     kv_small = 1 / x  # Small x: Kv(v, x) ~ 1 / x when v ~ 0

#     # Combine based on thresholds
#     kv_approx = torch.where(large_x, kv_large, kv_small)
#     return kv_approx


# def matern_covariance(d, alpha, sigma2, beta):
#     """
#     Matérn covariance function implemented with PyTorch.
#     Args:
#         d (torch.Tensor): Distance matrix (NxN or grid distance matrix).
#         alpha (float): Smoothness parameter.
#         sigma2 (float): Variance (amplitude) parameter.
#         beta (float): Scale parameter (controls correlation decay).
#     Returns:
#         torch.Tensor: Covariance matrix.
#     """
#     # Convert constants to tensors
#     alpha_tensor = torch.tensor(alpha, dtype=d.dtype, device=d.device)
#     beta_tensor = torch.tensor(beta, dtype=d.dtype, device=d.device)

#     # Compute sqrt_term
#     sqrt_term = torch.sqrt(torch.tensor(2 * alpha, dtype=d.dtype, device=d.device)) * d / beta_tensor
#     sqrt_term = torch.clamp(sqrt_term, min=1e-10)  # Avoid division by zero

#     # Compute kv using approximation
#     kv_term = approximate_kv(alpha_tensor, sqrt_term)

#     # Compute gamma(alpha) using PyTorch
#     gamma_alpha = torch.exp(torch.special.gammaln(alpha_tensor))

#     # Compute Matérn covariance
#     cov = sigma2 * (2 ** (1 - alpha_tensor)) / gamma_alpha * (sqrt_term ** alpha_tensor) * kv_term
#     cov[d == 0] = sigma2  # Handle the diagonal (d=0)

#     # Add jitter for numerical stability
#     cov += torch.eye(cov.size(0), dtype=cov.dtype, device=cov.device) * 1e-6

#     # Validate positive-definiteness
#     eigenvalues = torch.linalg.eigvalsh(cov)
#     if torch.any(eigenvalues <= 0):
#         raise ValueError("Covariance matrix is not positive-definite. Adjust parameters or add more jitter.")
    
#     return cov

# def generate_grf(model_shape, alpha=1.5, sigma2=1.0, beta=10.0, device="cpu", seed=None):
#     """
#     Generate a 2D Gaussian Random Field (GRF) using Matérn covariance function.
#     Args:
#         model_shape (tuple): Size of the output grid (h, w).
#         alpha (float): Smoothness parameter of the Matérn covariance.
#         sigma2 (float): Variance (amplitude) parameter.
#         beta (float): Scale parameter (correlation length).
#         device (str): Device for computation ("cpu" or "cuda").
#         seed (int, optional): Random seed for reproducibility.
#     Returns:
#         torch.Tensor: 2D Gaussian Random Field of size (h, w).
#     """
#     if seed is not None:
#         torch.manual_seed(seed)

#     h, w = model_shape
#     y = torch.arange(0, h, device=device)
#     x = torch.arange(0, w, device=device)
#     yy, xx = torch.meshgrid(y, x, indexing='ij')

#     # Compute pairwise distance matrix (flattened grid)
#     grid_points = torch.stack([yy.flatten(), xx.flatten()], dim=1).float()
#     distances = torch.cdist(grid_points, grid_points, p=2)
#     distances = distances.reshape(h * w, h * w)

#     # Compute covariance matrix using Matérn function
#     cov_matrix = matern_covariance(distances, alpha, sigma2, beta)

#     # Perform Cholesky decomposition to generate correlated samples
#     L = torch.linalg.cholesky(cov_matrix + 1e-6 * torch.eye(cov_matrix.size(0), device=device))  # Add jitter for stability
#     z = torch.randn(h * w, 1, device=device)  # Standard normal samples
#     grf = (L @ z).reshape(h, w)  # Correlated samples reshaped to grid

#     return grf